#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Create a pool xml catalog slice for the specified LFNs."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript

__RCSID__ = "$Id$"


def __getLfnsFromFile(optFiles, gaudiVerbose):
  import tempfile
  _a, tmpFile = tempfile.mkstemp(suffix=".py")
  runOpts = ""
  for opt in optFiles:
    if not os.path.exists(opt):
      gLogger.always("File not found: ", opt)
      DIRAC.exit(1)
    runOpts += opt + " "

  gaudiRun = "gaudirun.py -n -o %s %s" % (tmpFile, runOpts)
  if not gaudiVerbose:
    gaudiRun += " &>/dev/null"
  gLogger.info("Extract list of input files from", optFiles)

  gLogger.info("lb-run LHCb for getting environment")
  command = "lb-run --siteroot=/cvmfs/lhcb.cern.ch/lib LHCb/latest " + gaudiRun
  rc = os.system(command)
  if rc:
    gLogger.always("Error when parsing options files", optFiles)
    DIRAC.exit(rc)

  optDict = eval(open(tmpFile, 'r').read())
  os.remove(tmpFile)
  appInput = optDict.get('EventSelector', {}).get('Input')
  if not appInput:
    gLogger.always("Options file do not set EventSelector().Input")
    DIRAC.exit(1)

  return [inp.split()[0].split("'")[1].replace('LFN:', '') for inp in appInput]


def execute():
  from DIRAC.Core.Base import Script
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

  catalog = 'pool_xml_catalog.xml'
  depth = 1
  optFiles = []
  newOptFile = ''
  ignore = False
  verbose = False
  gaudiVerbose = False

  dmScript = DMScript()
  dmScript.registerSiteSwitches()
  dmScript.registerFileSwitches()
  Script.registerSwitch('', 'Options=', '   List of option files to consider')
  Script.registerSwitch('', 'NewOptions=', '   Name of a new options file to be generated with LFNs (default: none)')
  Script.registerSwitch('', 'Catalog=', '   Catalog name (default: %s' % catalog)
  Script.registerSwitch('', 'Depth=', '   Depth for ancestor consideration (default: none)')
  Script.registerSwitch('v', 'Verbose', '   Verbose gLogger.always(out')
  Script.registerSwitch('', 'Ignore', '   Ignore missing files')
  Script.registerSwitch('', 'GaudiVerbose', '   Set Gaudi verbose when parsing option files')

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ... ' % Script.scriptName]))
  Script.parseCommandLine(ignoreErrors=True)

  from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
  try:
    site = DMSHelpers().getShortSiteNames()['CERN']
  except (AttributeError, KeyError):
    site = 'LCG.CERN.cern'

  t0 = time.time()
  switches = Script.getUnprocessedSwitches()
  for o, a in switches:
    if o in ("Options"):
      optFiles += a.split(',')
    elif o in ("Catalog"):
      catalog = a
      ext = os.path.splitext(catalog)
      if len(ext[1]) == 0 and '/dev/' not in catalog:
        catalog += os.path.extsep + "xml"
    elif o in ("Depth"):
      try:
        depth = int(a)
      except Exception:
        gLogger.fatal("Invalid depth, must be integer")
        DIRAC.exit(1)
    elif o in ("v", 'Verbose'):
      verbose = True
      gLogger.setLevel('Info')
    elif o in ("NewOptions"):
      newOptFile = a
      ext = os.path.splitext(newOptFile)
      if len(ext[1]) == 0:
        newOptFile += os.path.extsep + "py"
    elif o in ("Ignore"):
      ignore = True
    elif o in ("GaudiVerbose"):
      gaudiVerbose = True

  if depth <= 0:
    gLogger.fatal("Invalid ancestor depth, should be >= 1")
    Script.showHelp(exitCode=1)

  if newOptFile and len(optFiles) != 1:
    if optFiles:
      gLogger.fatal("Generating an options file is only valid with a _single_ file as input")
    else:
      gLogger.fatal("Generating an options file is only valid with an option file as input")
    Script.showHelp()
    DIRAC.exit(1)
  if newOptFile:
    if depth > 1:
      gLogger.warn("New options file required, depth ignored...")
    if ignore:
      gLogger.warn("New options file required, cannot ignore missing files")
      ignore = False
    depth = 1
    if os.path.realpath(newOptFile) == os.path.realpath(optFiles[0]):
      gLogger.fatal("Cannot write a new option file overwriting the original one, select a name different from",
                    newOptFile)
      Script.showHelp(exitCode=1)

  sites = dmScript.getOption('Sites', [])
  if sites:
    if len(sites) != 1:
      gLogger.fatal("Provide at most one site...")
      Script.showHelp(exitCode=1)
    site = sites[0]

  if optFiles:
    lfnList = __getLfnsFromFile(optFiles, gaudiVerbose)
  else:
    lfnList = dmScript.getOption('LFNs', [])
  if not lfnList:
    gLogger.fatal("No option files and no list of files given")
    Script.showHelp(exitCode=1)

  rc = 0
  savedLevel = gLogger.getLevel()
  try:
      # Verify the user has a valid proxy
    done = 1
    while done and os.system("dirac-proxy-info --checkvalid > /dev/null") != 0:
      gLogger.always("You don't have a valid proxy, we create one...")
      done = os.system("lhcb-proxy-init")

    if depth > 1:
      from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
      bk = BookkeepingClient()
      result = bk.getFileAncestors(lfnList, depth)
      if not result['OK']:
        gLogger.fatal("Error getting ancestor files...")
        DIRAC.exit(1)
      lfnList = [anc['FileName'] for ancestors in result['Value']['Successful'].values()
                 for anc in ancestors] + list(result['Value']['Successful'])

    from DIRAC.Interfaces.API.Dirac import Dirac
    if not verbose:
      gLogger.setLevel("Exception")
    if newOptFile:
      catalog = "/dev/null"
    gLogger.info("List of LFNs:", lfnList)
    gLogger.info("Site:", site)
    gLogger.info("Catalog file:", catalog)
    result = Dirac().getInputDataCatalog(lfnList, site, catalog, ignoreMissing=ignore)
    if result["OK"]:
      result = result['Value']
      if result['Failed']:
        gLogger.always('Only a fraction of the input files are present and available at %s (%d missing)' %
                       (site, len(result['Failed'])))
      if not result['Successful']:
        gLogger.fatal('... and none was actually available!')
        DIRAC.exit(1)
      if newOptFile:
        lfnToPfn = result["Successful"]
        optLines = open(optFiles[0]).readlines()
        fo = open(newOptFile, 'w')
        for opt in optLines:
          for lfn in lfnList:
            if lfn in opt:
              # get the tURL for this lfn
              if lfn in lfnToPfn:
                pfn = lfnToPfn[lfn][0]['turl']
                opt = opt.replace("LFN:", '').replace(lfn, pfn)
              else:
                gLogger.error("No tURL found for ", lfn)
              break
          fo.write(opt)
        fo.close()
        gLogger.always("New options file %s successfully created" % newOptFile)
      else:
        gLogger.always("POOL XML catalog %s successfully created" % catalog)
        if os.path.exists(catalog + '.temp'):
          os.remove(catalog + '.temp')
        if os.path.isfile(catalog):
          catOption, ext = os.path.splitext(catalog)
          catOption += os.path.extsep + "py"
          fo = open(catOption, 'w')
          fo.write("FileCatalog().Catalogs = [ 'xmlcatalog_file:" + catalog + "' ]\n")
          fo.close()
          gLogger.always("==> You must add %s to your list of options file" % catOption)
    else:
      gLogger.error("Error getting the list of PFNs:", result['Message'])
      rc = 1
  except Exception as e:
    gLogger.exception("Exception caught while creating catalog or option file:", '', e)
    rc = 1

  gLogger.setLevel(savedLevel)
  gLogger.info("Total execution time for %d files: %5.2f seconds" % (len(lfnList), time.time() - t0))
  DIRAC.exit(rc)


@DIRACScript()
def main():
  execute()
  DIRAC.exit(0)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for Workflow Modules utilities."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
import copy
import os

from mock import MagicMock, patch

from DIRAC import gLogger

from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock
from LHCbDIRAC.Workflow.Modules.mock_Commons import prod_id, prod_job_id, wms_job_id, \
    workflowStatus, stepStatus, step_id, step_number,\
    step_commons, wf_commons

from LHCbDIRAC.Workflow.Modules.GaudiApplicationScript import GaudiApplicationScript
from LHCbDIRAC.Workflow.Modules.RootApplication import RootApplication
from LHCbDIRAC.Workflow.Modules.LHCbScript import LHCbScript
from LHCbDIRAC.Workflow.Modules.ErrorLogging import ErrorLogging


class ModulesApplicationsTestCase(unittest.TestCase):
  """Base class for the Modules Applications test cases."""

  def setUp(self):

    gLogger.setLevel('DEBUG')
    self.maxDiff = None

  def tearDown(self):

    for fileProd in ['prodConf_someApp_123_00000456_123_00000456_321.py', 'appLog', 'gaudi_extra_options.py',
                     'applicationError.txt', 'someApp', 'applicationLog.txt']:
      try:
        os.remove(fileProd)
      except OSError:
        continue


#############################################################################
# GaudiApplicationScript.py
#############################################################################

class GaudiApplicationScriptSuccess(ModulesApplicationsTestCase):

  @patch("LHCbDIRAC.Workflow.Modules.GaudiApplicationScript.RunApplication", side_effect=MagicMock())
  @patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  def test_execute(self, _patch, _patched):

    gas = GaudiApplicationScript(bkClient=bkc_mock, dm=dm_mock)
    gas.jobType = 'user'

    # no errors, no input data
    for wf_cs in copy.deepcopy(wf_commons):
      for s_cs in step_commons:
        s_cs['script'] = 'cat'
        self.assertTrue(gas.execute(prod_id, prod_job_id, wms_job_id,
                                    workflowStatus, stepStatus,
                                    wf_cs, s_cs,
                                    step_number, step_id)['OK'])

#############################################################################
# LHCbScript.py
#############################################################################


class LHCbScriptSuccess(ModulesApplicationsTestCase):

  def test_execute(self):

    lhcbScript = LHCbScript()

    lhcbScript.jobType = 'merge'
    lhcbScript.stepInputData = ['foo', 'bar']

    lhcbScript.production_id = prod_id
    lhcbScript.prod_job_id = prod_job_id
    lhcbScript.jobID = wms_job_id
    lhcbScript.workflowStatus = workflowStatus
    lhcbScript.stepStatus = stepStatus
    lhcbScript.workflow_commons = wf_commons
    lhcbScript.step_commons = step_commons[0]
    lhcbScript.step_number = step_number
    lhcbScript.step_id = step_id
    lhcbScript.executable = 'ls'
    lhcbScript.applicationLog = 'applicationLog.txt'

    # no errors, no input data
    for wf_cs in copy.deepcopy(wf_commons):
      for s_cs in step_commons:
        lhcbScript.workflow_commons = wf_cs
        lhcbScript.step_commons = s_cs
        lhcbScript._setCommand()
        lhcbScript._executeCommand()


class LHCbScriptFailure(ModulesApplicationsTestCase):

  def test_execute(self):

    lhcbScript = LHCbScript()

    lhcbScript.jobType = 'merge'
    lhcbScript.stepInputData = ['foo', 'bar']

    lhcbScript.production_id = prod_id
    lhcbScript.prod_job_id = prod_job_id
    lhcbScript.jobID = wms_job_id
    lhcbScript.workflowStatus = workflowStatus
    lhcbScript.stepStatus = stepStatus
    lhcbScript.workflow_commons = wf_commons
    lhcbScript.step_commons = step_commons[0]
    lhcbScript.step_number = step_number
    lhcbScript.step_id = step_id

    # no errors, no input data
    for wf_cs in copy.deepcopy(wf_commons):
      for s_cs in step_commons:
        lhcbScript.workflow_commons = wf_cs
        lhcbScript.step_commons = s_cs
        res = lhcbScript.execute()
        self.assertFalse(res['OK'])

#############################################################################
# RootApplication.py
#############################################################################


class RootApplicationSuccess(ModulesApplicationsTestCase):

  @patch("LHCbDIRAC.Workflow.Modules.RootApplication.RunApplication", side_effect=MagicMock())
  @patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  def test_execute(self, _patch, _patched):

    with open('someApp', 'w') as fd:
      fd.write('pippo')

    ra = RootApplication(bkClient=bkc_mock, dm=dm_mock)
    ra.applicationName = 'aRoot.py'
    ra.applicationVersion = 'v1r1'
    ra.rootType = 'py'
    ra.jobType = 'user'

    # no errors, no input data
    for wf_cs in copy.deepcopy(wf_commons):
      for s_cs in step_commons:
        self.assertTrue(ra.execute(prod_id, prod_job_id, wms_job_id,
                                   workflowStatus, stepStatus,
                                   wf_cs, s_cs,
                                   step_number, step_id)['OK'])


#############################################################################
# ErrorLogging.py
#############################################################################

class ErrorLoggingSuccess(ModulesApplicationsTestCase):

  @patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  def test_excecute(self, _patch):

    er = ErrorLogging()
    er.jobType = 'user'

    # no errors, no input data
    for wf_cs in copy.deepcopy(wf_commons):
      for s_cs in step_commons:
        self.assertTrue(er.execute(prod_id, prod_job_id, wms_job_id,
                                   workflowStatus, stepStatus,
                                   wf_cs, s_cs,
                                   step_number, step_id)['OK'])


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(ModulesApplicationsTestCase)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(GaudiApplicationScriptSuccess))
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(LHCbScriptSuccess))
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(LHCbScriptFailure))
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(RootApplicationSuccess))
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ErrorLoggingSuccess))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.ResourceStatusSystem.Policy.Configurations.

POLICIESMETA_LHCB policies POLICIESMETA
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.ResourceStatusSystem.Policy.Configurations import POLICIESMETA

# LHCb Policies

POLICIESMETA_LHCB = {'TransferQualitySource': {'description': 'Transfers from element, quality measure',
                                               'module': 'TransferQualityPolicy',
                                               'command': ('TransferCommand', 'TransferCommand'),
                                               'args': {'direction': 'Source',
                                                        'metric': 'Quality',
                                                        'hours': 2}},
                     'TransferQualityDestination': {'description': 'Transfers to element, quality measure',
                                                    'module': 'TransferQualityPolicy',
                                                    'command': ('TransferCommand', 'TransferCommand'),
                                                    'args': {'direction': 'Destination', 'metric':
                                                              'Quality', 'hours': 2}},
                     }

################################################################################
# Old stuff

policies = {'DTScheduled': {'description': 'Ongoing and scheduled down-times',
                            'module': 'DTPolicy',
                            'commandInNewRes': ('GOCDBStatusCommand', 'GOCDBStatusCommand'),
                            # 'command'         : ( 'GOCDBStatusCommand', 'DTCachedCommand' ),
                            'command': ('GOCDBStatusCommand', 'GOCDBStatusCommand'),
                            'args': {'hours': 12},  # Fix to avoid querying the CS on load time, to be fixed
                            'Site_Panel': [{'WebLink': {'CommandIn': ('GOCDBStatusCommand',
                                                                      'DTInfoCachedCommand'),
                                                        'args': None}}],
                            'Resource_Panel': [{'WebLink': {'CommandIn': ('GOCDBStatusCommand',
                                                                          'DTInfoCachedCommand'),
                                                            'args': None}}]},

            'OnStorageElementPropagation': {'description': 'How the storage element\'s nodes are behaving in the RSS',
                                            'module': 'DownHillPropagationPolicy',
                                            'command': ('RSCommand', 'MonitoredStatusCommand'),
                                            'args': ('Resource', ),
                                            'SE_Panel': [{'RSS': 'ResOfStorEl'}]},
            'TransferQuality': {'description': 'SE transfer quality',
                                'module': 'TransferQualityPolicy',
                                'commandInNewRes': ('DIRACAccountingCommand', 'TransferQualityCommand'),
                                'argsNewRes': None,
                                'command': ('DIRACAccountingCommand', 'TransferQualityFromCachedPlotCommand'),
                                'args': ('DataOperation', 'TransferQualityByDestSplitted_2'),

                                'SE_Panel': [{'FillChart - Transfer quality in the last 24 hours':
                                              {'CommandIn': ('DIRACAccountingCommand', 'CachedPlotCommand'),
                                               'args': ('DataOperation', 'TransferQualityByDestSplitted_24'),
                                               'CommandInNewRes': ('DIRACAccountingCommand', 'DIRACAccountingCommand'),
                                               'argsNewRes': ('DataOperation', 'Quality',
                                                              {'Format': 'LastHours', 'hours': 24},
                                                              'Destination',
                                                              {'OperationType': 'putAndRegister'})}}, ]}}

# Update DIRAC policies with LHCbDIRAC policies
POLICIESMETA.update(POLICIESMETA_LHCB)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# Test for LogErr.py

import json
import ast
import os
import pytest

# sut
from LHCbDIRAC.Core.Utilities.LogErr import createJSONtable


# Define test data
jsonDataMultiple = [
    {
        'G4Exception': [
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''}
        ]
    },
    {
        'ERROR Gap not found!': [
            {'runnr': '  Run 133703', 'eventnr': 'Evt 29'}
        ]
    },
    {
        'The signal decay mode is not defined in the main DECAY.DEC table': [
            {'runnr': '  Run 133703', 'eventnr': 'Evt 29'},
            {'runnr': '  Run 123', 'eventnr': 'Evt 30'}
        ]
    },
    {
        'G4Exception : StuckTrack': [
            {'runnr': '  Run 133703', 'eventnr': 'Evt 29'},
            {'runnr': '  Run 123', 'eventnr': 'Evt 30'},
            {'runnr': '  Run 1234', 'eventnr': 'Evt 31'}
        ]
    },
    {
        'G4Exception : 001': [
            {'runnr': '  Run 133703', 'eventnr': 'Evt 29'},
            {'runnr': '  Run 123', 'eventnr': 'Evt 30'},
            {'runnr': '  Run 1234', 'eventnr': 'Evt 31'}
        ]
    }
]

expectedMultiple = json.dumps(
    {
        "ERROR Gap not found!": 1,
        "wmsID": "5",
        "ProductionID": "4",
        "JobID": "3",
        "G4Exception": 10,
        "The signal decay mode is not defined in the main DECAY.DEC table": 2,
        "G4Exception : StuckTrack": 3,
        "G4Exception : 001": 3
    }, indent=2)


jsonDataSingle = [
    {
        'G4Exception : InvalidSetup': [
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''},
            {'runnr': '', 'eventnr': ''}
        ]
    }
]

expectedSingle = json.dumps(
    {
        "G4Exception : InvalidSetup": 10,
        "wmsID": "5",
        "ProductionID": "4",
        "JobID": "3"
    }, indent=2)

jsonDataEmpty = []

expectedEmpty = json.dumps(
    {
        "wmsID": "5",
        "ProductionID": "4",
        "JobID": "3"
    }, indent=2)


name = 'test_createJSONtable.json'


@pytest.mark.parametrize("input, expected", [
    (jsonDataMultiple, expectedMultiple),
    (jsonDataSingle, expectedSingle),
    (jsonDataEmpty, expectedEmpty)
])
def test_createJson(input, expected):
  createJSONtable(input, name, '3', '4', '5')
  with open(name, 'r') as f:
    fileOutput = f.read()

  # Convert to dict()
  fileOutput = ast.literal_eval(fileOutput)
  expected = ast.literal_eval(expected)

  assert fileOutput == expected
  os.remove(name)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve from Bookkeeping the number of Jobs at each Site for a given Production."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ... ProdID' % Script.scriptName,
      'Arguments:',
      '  ProdID:   Production ID (mandatory)']))
  Script.parseCommandLine(ignoreErrors=True)
  args = Script.getPositionalArgs()
  print("args =", args)
  if len(args) < 1:
    Script.showHelp()

  exitCode = 0

  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  bk = BookkeepingClient()
  try:
    prod = int(args[0])
  except BaseException:
    Script.showHelp(exitCode=1)

  res = bk.getNbOfJobsBySites(prod)

  if res['OK']:
    if not res['Value']:
      print("No jobs for production", prod)
      DIRAC.exit(0)
    sites = dict([(site, num) for num, site in res['Value']])
    shift = 0
    for site in sites:
      shift = max(shift, len(site) + 2)
    print('Site Name'.ljust(shift), 'Number of jobs')
    for site in sorted(sites):
      print(site.ljust(shift), str(sites[site]))
  else:
    print("ERROR getting number of jobs for %s:" % str(prod), res['Message'])
    exitCode = 2


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Report sisters or cousins (i.e. descendant of a parent or ancestor) for a (list of) LFN(s)"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, Script

  dmScript = DMScript()
  dmScript.registerFileSwitches()
  Script.registerSwitch('', 'All', 'Do not restrict to sisters with replicas')
  Script.registerSwitch('', 'Full', 'Get full metadata information on sisters')
  level = 1
  Script.registerSwitch(
      '',
      'Depth=',
      'Number of ancestor levels (default: %d), 2 would be cousins, 3 grand-cousins etc...' %
      level)
  Script.registerSwitch('', 'Production=', 'Production to check for sisters (default=same production)')
  Script.registerSwitch('', 'AllFileTypes', 'Consider also files with a different type')
  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ... [LFN|File] [Level]' % Script.scriptName]))

  Script.parseCommandLine(ignoreErrors=True)

  from LHCbDIRAC.BookkeepingSystem.Client.ScriptExecutors import executeFileSisters
  executeFileSisters(dmScript, level=level)


if __name__ == "__main__":
  main()



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""This script is used to flag a given files which belongs a certain run.

1. We flag OK or BAD without specifying a processing pass: this flags the RAW and all derived files as OK or BAD
2. If a processing pass is specified (and we allow one PP at a time only):
     2.1 For BAD, only that processing pass (and derived) is flagged BAD.
         RAW is left unchanged unless PP is '/Real Data' which is then similar to 1
     2.2 For OK:
               - if '/Real Data', only the RAW are flagged OK.
                 Derived data are left unchanged. If one wants to flag everything, use 1.
               - else that processing pass (and derived) and the RAW are flagged OK
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

import DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


def checkDQFlag(dqFlag):
  """Make sure the DQ flag is a known one."""
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

  res = BookkeepingClient().getAvailableDataQuality()
  if not res['OK']:
    return res

  if dqFlag in res['Value']:
    return S_OK()
  else:
    return S_ERROR("Data quality flag is not in the db")


def flagFileList(filename, dqFlag):
  """Flag a LFN or a list of LFN contained in a file."""
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

  lfns = []

  # Load the list of LFN
  try:
    with open(filename) as f:
      for lfn in f:
        lfns.append(lfn.strip())
  except IOError:
    lfns = filename.split(',')

  # Now flag the LFN
  res = BookkeepingClient().setFileDataQuality(lfns, dqFlag)

  if not res['OK']:
    return res
  else:
    gLogger.notice('The data quality has been set %s for %d files:' % (dqFlag, len(lfns)))

  return S_OK()


def flagRun(runNumber, procPass, dqFlag, flagRAW=False):
  """Flag a run given its number, the processing pass and the DQ flag."""
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

  res = getProcessingPasses(runNumber, '/Real Data')
  if not res['OK']:
    return S_ERROR('flagRun: %s' % res['Message'])

  allProcessingPasses = set(res['Value'])

  gLogger.notice("Available processing passes:", str(allProcessingPasses))

  processingPasses = []
  if procPass:  # we will flag a certain processing: for example: /Real Data/Reco09
    for processingPass in allProcessingPasses:
      if processingPass.startswith(procPass):
        processingPasses.append(processingPass)
  else:  # we will flag everything
    processingPasses = allProcessingPasses

  if flagRAW:
    processingPasses.append('/Real Data')

  if not processingPasses:
    return S_ERROR('%s is not a valid processing pass.' % procPass)

    # Flag the processing passes
  for processingPass in processingPasses:
    res = BookkeepingClient().setRunAndProcessingPassDataQuality(
        runNumber,
        processingPass,
        dqFlag,
    )
    if not res['OK']:
      return S_ERROR('flagRun: processing pass %s\n error: %s' % (processingPass, res['Message']))
    else:
      gLogger.notice('Run %d Processing Pass %s flagged %s' % (runNumber,
                                                               processingPass,
                                                               dqFlag))

  return S_OK()


def getProcessingPasses(runNumber, procPass):
  """Find all known processing passes for the selected configurations."""
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

  res = BookkeepingClient().getRunConfigurationsAndDataTakingCondition(int(runNumber))
  if not res['OK']:
    return res

  bkDict = res['Value']
  bkDict['RunNumber'] = runNumber

  passes = []
  res = browseBkkPath(bkDict, procPass, passes)
  if not res['OK']:
    return res
  return S_OK(passes)


def browseBkkPath(bkDict, processingPass, visitedProcessingPass):
  """This method visit the processing passes started from processingPass.

  The visited processing passes are kept in visitedProcessingPass
  """
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

  res = BookkeepingClient().getProcessingPass(bkDict, processingPass)
  if not res['OK']:
    gLogger.error('Cannot load the processing passes for head % in Version %s Data taking condition %s' %
                  (processingPass, bkDict['ConfigVersion'], bkDict['ConditionDescription']))
    gLogger.error(res['Message'])
    return res

  records = res['Value'][0]  # 0 contains the processing passes 1 contains the event types
  if 'Name' in records['ParameterNames']:  # this mean we have processing passes
    # this is the name of the processing pass: 'ParameterNames': ['Name']
    index = records['ParameterNames'].index('Name')
    passes = sorted([os.path.join(processingPass, record[index]) for record in records['Records']])
  else:
    passes = []

  if passes:
    for pName in passes:
      visitedProcessingPass.append(pName)
      browseBkkPath(bkDict, pName, visitedProcessingPass)

  return S_OK()


@DIRACScript()
def main():
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

  Script.registerSwitch("l:", "lfnfile=", "Flag a LFN or list of LFN")
  Script.registerSwitch("r:", "run=", "Flag a run")
  Script.registerSwitch("p:", "processingPass=", "Processing pass for which a run should be flagged")
  Script.registerSwitch("q:", "dataqualityflag=", "Data quality flag")

  Script.setUsageMessage(__doc__ + '\n'.join([
      '\nArguments:',
      '  lfnfile (string) : [LFN|filename]',
      '  processingPass (string) : porcessing pass(es)',
      '  runNumber (int): run number to be flagged',
      '  dataqualityflag (string): data quality\n', ]))

  Script.parseCommandLine(ignoreErrors=True)

  exitCode = 0

  params = {'lfn': None,
            'runnumber': None,
            'dqflag': None,
            'processingpass': None}

  for switch in Script.getUnprocessedSwitches():
    if switch[0].lower() in ('l', 'lfnfile'):
      params['lfn'] = switch[1]
    elif switch[0].lower() in ('r', 'run'):
      params['runnumber'] = int(switch[1])
    elif switch[0].lower() in ('q', 'dataqualityflag'):
      params['dqflag'] = switch[1]
    elif switch[0].lower() in ('p', 'processingPass'):
      params['processingpass'] = switch[1]

  if params['lfn'] is None and params['runnumber'] is None:
    gLogger.fatal('Please specify run number or an lfn!')
    DIRAC.exit(1)
  if params['dqflag'] is None:
    gLogger.fatal('Please specify the data quality flag!')
    DIRAC.exit(1)

  processingPass = None
  if params['processingpass']:
    if len(params['processingpass'].split(',')) > 1:
      gLogger.fatal("More than one processing pass is given. Please use only one processing pass!")
      DIRAC.exit(1)
    else:
      processingPass = params['processingpass']
      if not processingPass.startswith('/'):
        processingPass = '/' + processingPass
      processingPass = processingPass.replace('RealData', 'Real Data')  # To be consistent with other BK scripts ;-)
      if not processingPass.startswith('/Real Data'):
        processingPass = '/Real Data' + processingPass

  res = checkDQFlag(params['dqflag'])
  if not res['OK']:
    gLogger.fatal("%s - %s" % (params['dqflag'], res['Message']))
    DIRAC.exit(1)

  if params['lfn']:
    res = flagFileList(params['lfn'], params['dqflag'])
    if res['OK']:
      gLogger.notice("Files are flagged!")
      DIRAC.exit(exitCode)
    else:
      gLogger.fatal(res['Message'])

  if params['runnumber']:
    if not processingPass:  # processing pass is not given
      # this flags the RAW and all derived files as the given dq flag
      res = flagRun(
          runNumber=params['runnumber'],
          procPass='/',
          dqFlag=params['dqflag'],
          flagRAW=True
      )
      if not res['OK']:
        gLogger.fatal(res['Message'])
        DIRAC.exit(1)
    else:  # the processing pass is given
      if params['dqflag'] == 'BAD':  # only that processing pass (and derived) is flagged BAD. RAW is left unchanged
        res = flagRun(
            runNumber=params['runnumber'],
            procPass=processingPass,
            dqFlag=params['dqflag'],
        )
        if not res['OK']:
          gLogger.fatal(res['Message'])
          DIRAC.exit(1)
      elif params['dqflag'] == 'OK':
        if processingPass == '/Real Data':  # only flag the RAW
          res = BookkeepingClient().setRunAndProcessingPassDataQuality(
              params['runnumber'],
              '/Real Data',
              params['dqflag'],
          )
          if not res['OK']:
            gLogger.fatal(res['Message'])
            DIRAC.exit(1)
          else:
            gLogger.notice("%d flagged OK" % params['runnumber'])
        else:
          res = flagRun(
              runNumber=params['runnumber'],
              procPass=processingPass,
              dqFlag=params['dqflag'],
              flagRAW=True,
          )
          if not res['OK']:
            gLogger.fatal(res['Message'])
            DIRAC.exit(1)

  DIRAC.exit(exitCode)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module holding MCStatsClient class."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from DIRAC.Core.Base.Client import Client, createClient


@createClient('ProductionManagement/MCStatsElasticDB')
class MCStatsClient(Client):
  """Client for MCStatsElasticDB
  """

  def __init__(self, **kwargs):
    """simple constructor."""

    super(MCStatsClient, self).__init__(**kwargs)
    self.setServer('ProductionManagement/MCStatsElasticDB')

  def set(self, typeName, data):
    """set some data in a certain type.

    :params str typeName: type name (e.g. 'gaussErrors')
    :params dict data: dictionary inserted

    :returns: S_OK/S_ERROR
    """
    return self._getRPC().set(typeName, data)

  def get(self, typeName, productionID):
    """get per Job ID.

    :params str typeName: type name (e.g. 'gaussErrors')
    :params int productionID: production ID
    """
    return self._getRPC().get(typeName, productionID)

  def remove(self, typeName, productionID):
    """remove data for productionID.

    :params str typeName: type name (e.g. 'gaussErrors')
    :params int productionID: production ID
    """
    return self._getRPC().remove(typeName, productionID)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_BKK_Client_BaseESManager."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
import LHCbDIRAC.BookkeepingSystem.Client.BaseESManager as moduleTested

__RCSID__ = "$Id$"


class BaseESManager_TestCase(unittest.TestCase):

  def setUp(self):
    """Setup."""

    self.testClass = moduleTested.BaseESManager

  def tearDown(self):
    """TearDown."""

    del self.testClass


class BaseESManager_Success(BaseESManager_TestCase):

  def test_instantiate(self):
    """tests that we can instantiate one object of the tested class."""
    client = self.testClass()
    self.assertEqual('BaseESManager', client.__class__.__name__)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: StorageUsageClient

.. module: StorageUsageClient

:synopsis: Lightweight possbile client to the StorageUsageDB.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Base.Client import Client, createClient


@createClient('DataManagement/StorageUsage')
class StorageUsageClient(Client):
  """
  .. class:: StorageUsageClient
  """

  def __init__(self, url=None, **kwargs):
    """c'tor."""
    super(StorageUsageClient, self).__init__(**kwargs)
    if url:
      self.setServer(url)
    self.setServer('DataManagement/StorageUsage')



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Extension of DIRAC SiteDirector.

Simply defines what to send.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.WorkloadManagementSystem.Agent.SiteDirector import SiteDirector as DIRACSiteDirector


class SiteDirector(DIRACSiteDirector):
  """Simple extension of the DIRAC site director to send LHCb specific pilots
  (with a custom list of commands)"""

  def beginExecution(self):
    """just simple redefinition."""
    res = DIRACSiteDirector.beginExecution(self)
    if not res['OK']:
      return res

    self.lbRunOnly = self.am_getOption('lbRunOnly', False)

    return S_OK()

  def _getPilotOptions(self, queue):
    """Adding LHCb specific options."""
    pilotOptions = super(SiteDirector, self)._getPilotOptions(queue)

    if self.lbRunOnly:
      pilotOptions.append('-o lbRunOnly')

    return pilotOptions



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""An agent to check for MCSimulation productions that have undergone the
testing phase. Productions that have the status Idle and are also in the table
StoredJobDescription have undergone testing. A report is created by the agent
from the results of the test phase and emailed to the Production Manager.

Author: Simon Bidwell
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR

from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Workflow.Workflow import fromXMLString
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getUserOption

from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Client.Production import Production
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.Workflow.Modules.ModulesUtilities import getEventsToProduce, getCPUNormalizationFactorAvg

AGENT_NAME = 'Transformation/MCSimulationTestingAgent'


class MCSimulationTestingAgent (AgentModule):
  """An agent to check for MCSimulation productions that have undergone the
  testing phase.

  Productions that have the status Idle and are also in the table
  StoredJobDescription have undergone testing. A report is created by
  the agent from the results of the test phase and emailed to the
  Production Manager
  """

  def __init__(self, *args, **kwargs):
    """c'tor."""
    AgentModule.__init__(self, *args, **kwargs)
    self.transClient = None
    self.bkClient = None
    self.notifyClient = None
    self.operations = None

    self.failedTransIDs = []

  def initialize(self):
    self.transClient = TransformationClient()
    self.bkClient = BookkeepingClient()
    self.notifyClient = NotificationClient()
    self.operations = Operations()

    self.email = self.am_getOption("MailTo", '')

    return S_OK()

  def execute(self):
    # get all the idle transformations
    extendableTTypes = Operations().getValue('Transformations/ExtendableTransfTypes', ['MCSimulation'])
    res = self.transClient.getTransformations(condDict={"Status": "Idle", "Type": extendableTTypes})
    if res['OK']:
      idleTransformations = res['Value']
      idleTransformations = [d.get("TransformationID") for d in idleTransformations]
      self.log.verbose("Found %d Idle MC transformations" % len(idleTransformations))
      self.log.debug("Idle transformations found: %s" % ','.join([str(it) for it in idleTransformations]))
    else:
      self.log.error("Call to Transformation Client service failed", res['Message'])
      return res

    # get all the IDs of transformations undergoing a testing phase
    res = self.transClient.getStoredJobDescriptionIDs()
    if res['OK']:
      testingSimulations = res['Value']
      testingSimulations = [pair[0] for pair in testingSimulations]
      self.log.verbose("Found %d MC transformations undergoing a testing phase" % len(testingSimulations))
      self.log.debug("MC transformations found undergoing a testing phase: %s" %
                     ','.join([str(ts) for ts in testingSimulations]))
    else:
      self.log.error("Call to Transformation Client service failed", res['Message'])
      return res

    # get the IDs that occur in both idle transformations and testing phase
    idleSimulations = list(set(testingSimulations).intersection(idleTransformations))
    # remove those that we know failed
    idleSimulations = list(set(idleSimulations).difference(self.failedTransIDs))
    self.log.info("MC transformations under considerations: %s (will loop on them)" %
                  ','.join([str(idS) for idS in idleSimulations]))
    for transID in idleSimulations:
      self.log.info("Looking into %d" % transID)
      tasks = self.transClient.getTransformationTasks(condDict={"TransformationID": transID})
      if not tasks['OK']:
        self.log.error("Call to Transformation Client service failed", tasks['Message'])
        continue
      else:
        tasks = tasks['Value']
        numberOfTasks = len(tasks)
        numberOfDoneTasks = sum(1 for d in tasks if d.get("ExternalStatus") == "Done")
        self.log.verbose(
            "TransID = %d, numberOfTasks = %d, numberOfDoneTasks = %d" %
            (transID, numberOfTasks, numberOfDoneTasks))
        if numberOfTasks == numberOfDoneTasks:
          self.log.info("All tasks have passed so the request can be accepted and the transformation updated")
          res = self._activateTransformation(transID, tasks)
          if not res['OK']:
            self.log.error("Error Activating Production", res['Message'])
        else:
          self.log.warn("There are failed tasks")
          report = self.__createReport(tasks)
          numberOfFailedTasks = sum(1 for d in tasks if d.get('ExternalStatus') == 'Failed')
          if numberOfFailedTasks == numberOfTasks:
            # all tasks have failed so the request can be rejected and an email report sent
            self._sendReport(report)
            self.log.warn("Transformation " + str(transID) + " failed the testing phase")
            self.failedTransIDs.append(transID)
          else:
            # only some tasks have failed so continue but send a warn email
            self.log.warn("Transformation " + str(transID) + " failed partially the testing phase, continuing anyway")
            doneTasks = list()
            for d in tasks:
              if d.get("ExternalStatus") == "Done":
                doneTasks.append(d)
            if not doneTasks:
              self.log.info("No tasks done for Transformation %d" % transID)
              continue
            res = self._activateTransformation(transID, doneTasks)
            if not res['OK']:
              self.log.error("Error Activating Production", res['Message'])
              continue
            subject = "MCSimulation Test Failure Report. TransformationID: " + str(transID) + " - some tasks failed"
            report['subject'] = subject
            self._sendReport(report)

    return S_OK()

  def _activateTransformation(self, transID, tasks):
    """Calculate parameters, update the workflow, then move the production to
    Active."""
    parameters = self._calculateParameters(tasks)
    if not parameters['OK']:
      self.log.error("Error calculating parameters", parameters['Message'])
      return parameters
    else:
      parameters = parameters['Value']
      self.log.verbose("TransID = %d, Calculated Parameters: %s" % (transID, str(parameters)))
      workflow = self._updateWorkflow(transID, int(round(float(parameters['CPUe']))), parameters['MCCpu'])
      if workflow['OK']:
        workflow = workflow['Value']
        res = self._updateTransformationsTable(transID, workflow)
        if not res['OK']:
          self.log.error("Error updating transformations table", res['Message'])
          return res
        else:
          self.log.info("Transformation " + str(transID) + " passed the testing phase and is now set to active")

    return S_OK()

  def __createReport(self, tasks):
    """creates a report from a failed task to email to the production
    manager."""
    dateformat = '%d/%m/%Y %H:%M'
    transformationID = tasks[0]["TransformationID"]
    transformation = self.transClient.getTransformations(condDict={"TransformationID": transformationID})
    transformation = transformation['Value'][0]
    subject = "MCSimulation Test Failure Report. TransformationID: " + str(transformationID)
    body = [subject]
    body.append("")
    body.append("Transformation:")
    body.append("----------------------------------------------------------------------")
    body.append("TransformationID: " + str(transformation["TransformationID"]))
    body.append("TransformationName: " + transformation["TransformationName"])
    body.append("LastUpdate: " + transformation["LastUpdate"].strftime(dateformat))
    body.append("Status: " + transformation["Status"])
    body.append("Description: " + transformation["Description"])
    body.append("TransformationFamily: " + str(transformation["TransformationFamily"]))
    body.append("Plugin: " + transformation["Plugin"])
    body.append("Type: " + transformation["Type"])
    body.append("AgentType: " + transformation["AgentType"])
    body.append("GroupSize: " + str(transformation["GroupSize"]))
    body.append("MaxNumberOfTasks: " + str(transformation["MaxNumberOfTasks"]))
    body.append("AuthorDN: " + transformation["AuthorDN"])
    body.append("TransformationGroup: " + transformation["TransformationGroup"])
    body.append("InheritedFrom: " + str(transformation["InheritedFrom"]))
    body.append("CreationDate: " + transformation["CreationDate"].strftime(dateformat))
    body.append("FileMask: " + transformation["FileMask"])
    body.append("EventsPerTask: " + str(transformation["EventsPerTask"]))
    body.append("AuthorGroup: " + transformation["AuthorGroup"])
    body.append("")
    body.append("Number of Tasks: " + str(len(tasks)))
    body.append("Tasks:")
    body.append("----------------------------------------------------------------------")
    for task in tasks:
      body.append("TaskID: " + str(task['TaskID']))
      body.append("TargetSE: " + task['TargetSE'])
      body.append("LastUpdateTime: " + task['LastUpdateTime'].strftime(dateformat))
      body.append("RunNumber: " + str(task.get('RunNumber', 0)))
      body.append("CreationTime: " + task['CreationTime'].strftime(dateformat))
      body.append("ExternalID: " + str(task['ExternalID']))
      body.append("ExternalStatus: " + task['ExternalStatus'])
      body.append("")
    return {'subject': subject, 'body': body}

  def _sendReport(self, report):
    """sends a given report to the production manager."""
    if not self.email:
      self.email = getUserOption(self.operations.getValue("Shifter/ProductionManager/User"), 'Email')
    body = '\n'.join(report['body'])
    res = self.notifyClient.sendMail(
        self.email,
        report['subject'],
        body,
        self.email,
        localAttempt=False)
    if not res['OK']:
      self.log.error("sendMail failed", res['Message'])
    else:
      self.log.info('Mail summary sent to production manager')

  def _calculateParameters(self, tasks):
    """Calculates the CPU time per event for the production."""
    jobIds = [int(x['ExternalID']) for x in tasks]
    res = self.bkClient.bulkJobInfo({'jobId': jobIds})
    if not res['OK']:
      self.log.error("Error calling bkClient", res['Message'])
      return S_ERROR(res['Message'])
    successful = res['Value']['Successful']
    self.log.debug("Successful tasks: %s" % str(successful))
    if not successful:
      self.log.error("There are no successful tasks")
      return S_ERROR("There are no successful tasks")

    events = 0
    CPUeJobTotal = 0.0
    for job in successful.values():
      cpuJob = 0
      for bkJob in job:
        if bkJob['ApplicationName'] in ['Gauss', 'Boole', 'Moore', 'Brunel', 'DaVinci']:
          if not events:
            events = bkJob['NumberOfEvents']
          timeInSeconds = bkJob['CPUTIME']
          cpuJob += timeInSeconds * bkJob['WNCPUHS06']
      CPUeJob = cpuJob / events
      self.log.debug("CPUeJob = %d" % CPUeJob)

      CPUeJobTotal += CPUeJob

    CPUe = CPUeJobTotal / len(successful)
    # We want to produce at least 25 events per job...
    MCCpu = str(25 * int(round(float(CPUe))))
    self.log.verbose("CPUe = %d, MCCpu = %s" % (CPUe, MCCpu))
    return S_OK({'CPUe': CPUe, 'MCCpu': MCCpu})

  def _updateWorkflow(self, transID, CPUe, MCCpu):
    """Updates the workflow of a savedProductionDescription to reflect the
    calculated CPUe."""
    res = self.transClient.getStoredJobDescription(transID)
    if res['OK']:
      workflow = fromXMLString(res['Value'][0][1])
      prod = Production()
      prod.LHCbJob.workflow = workflow
      prod.setParameter('CPUe', 'string', str(CPUe), 'CPU time per event')
      prod.LHCbJob.setCPUTime(MCCpu)
      self.log.info("Transformation ", str(transID))
      self.log.info("Calculated CPUTime: ", str(CPUe))
      self.log.info("CpuTime: ", str(MCCpu))

      # maximum number of events to produce
      # try to get the CPU parameters from the configuration if possible
      cpuTimeAvg = Operations().getValue('Transformations/CPUTimeAvg')
      if cpuTimeAvg is None:
        self.log.info('Could not get CPUTimeAvg from config, defaulting to %d' % 200000)
        cpuTimeAvg = 200000

      try:
        CPUNormalizationFactorAvg = getCPUNormalizationFactorAvg()
      except RuntimeError:
        self.log.info('Could not get CPUNormalizationFactorAvg, defaulting to %f' % 1.0)
        CPUNormalizationFactorAvg = 1.0

      max_e = getEventsToProduce(CPUe, cpuTimeAvg, CPUNormalizationFactorAvg)
      prod.setParameter('maxNumberOfEvents', 'string', str(max_e), 'Maximum number of events to produce (Gauss)')
      return S_OK(prod.LHCbJob.workflow.toXML())
    else:
      self.log.error("Call to Transformation Client service failed", res['Message'])
      return res

  def _updateTransformationsTable(self, transID, workflow):
    """Puts the modified workflow from the savedProductionDescription table
    into the transformations table and removes it from the
    savedProductionDescription table."""
    transformation = self.transClient.getTransformations(condDict={"TransformationID": transID})
    if transformation['OK']:
      body = self.transClient.setTransformationParameter(transID, "Body", workflow)
      status = self.transClient.setTransformationParameter(transID, "Status", "Active")
      if body['OK'] and status['OK']:
        res = self.transClient.removeStoredJobDescription(transID)
        if not res['OK']:
          self.log.error("Call to removeStoredJobDescription failed", res['Message'])
          return res
        self.log.info("Transformation %s has an updated body and Status set to active" % transID)
        return S_OK()
      else:
        self.log.error("One of the updates has failed so set them both back to the previous value to ensure atomicity")
        self.log.debug(str(transformation['Value'][0]['Body']))
        res = self.transClient.setTransformationParameter(transID, "Body", transformation['Value'][0]['Body'])
        if not res['OK']:
          self.log.error("Failure calling setTransformationParameter", res['Message'])
          return res
        res = self.transClient.setTransformationParameter(transID, "Status", transformation['Value'][0]['Status'])
        if not res['OK']:
          self.log.error("Failure calling setTransformationParameter", res['Message'])
          return res
    else:
      self.log.error("Call to getTransformations failed", transformation['Message'])
      return transformation



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCb Bookkeeping database client."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from LHCbDIRAC.BookkeepingSystem.Client.BaseESClient import BaseESClient
from LHCbDIRAC.BookkeepingSystem.Client.LHCbBookkeepingManager import LHCbBookkeepingManager

__RCSID__ = "$Id$"


class LHCB_BKKDBClient(BaseESClient):
  """Client which used to browse the Entities."""
  #############################################################################

  def __init__(self, url=None, web=False, welcome=True):
    """Initialize the basic class."""
    BaseESClient.__init__(self, LHCbBookkeepingManager(url=url, web=web, welcome=welcome), '/')
  #############################################################################

  def get(self, path=""):
    """get path."""
    return self.getManager().get(path)

  #############################################################################
  def help(self):
    """help function."""
    return self.getManager().help()  # pylint: disable=no-member

  #############################################################################
  def getPossibleParameters(self):
    """available trees."""
    return self.getManager().getPossibleParameters()  # pylint: disable=no-member

  #############################################################################
  def setParameter(self, name):
    """tree used."""
    return self.getManager().setParameter(name)  # pylint: disable=no-member

  #############################################################################
  def getLogicalFiles(self):
    """lfns."""
    return self.getManager().getLogicalFiles()  # pylint: disable=no-member

  #############################################################################
  def getFilesPFN(self):
    """PFNS."""
    return self.getManager().getFilesPFN()  # pylint: disable=no-member

  #############################################################################
  def getNumberOfEvents(self, files):
    """number of events."""
    return self.getManager().getNumberOfEvents(files)  # pylint: disable=no-member

  #############################################################################
  def writeJobOptions(self, files, optionsFile="jobOptions.opts",
                      savedType=None, catalog=None, savePfn=None, dataset=None):
    """Gaudi card."""
    return self.getManager().writeJobOptions(files, optionsFile,  # pylint: disable=no-member
                                             savedType, catalog, savePfn, dataset)

  #############################################################################
  def getJobInfo(self, lfn):
    """how a file is created."""
    return self.getManager().getJobInfo(lfn)  # pylint: disable=no-member

  #############################################################################
  def setVerbose(self, value):
    """only important information."""
    return self.getManager().setVerbose(value)  # pylint: disable=no-member

  #############################################################################
  def setAdvancedQueries(self, value):
    """Advanced queries."""
    return self.getManager().setAdvancedQueries(value)  # pylint: disable=no-member

  #############################################################################
  def getLimitedFiles(self, selectionDict, sortDict, startItem, maxitems):
    """get files used by Web portal."""
    return self.getManager().getLimitedFiles(selectionDict, sortDict,  # pylint: disable=no-member
                                             startItem, maxitems)

  #############################################################################
  def getAncestors(self, files, depth):
    """ancestor of files."""
    return self.getManager().getAncestors(files, depth)  # pylint: disable=no-member

  #############################################################################
  def getLogfile(self, filename):
    """log file of a given file."""
    return self.getManager().getLogfile(filename)  # pylint: disable=no-member

  #############################################################################
  def writePythonOrJobOptions(self, startItem, maxitems, path, optstype):
    """python job option."""
    return self.getManager().writePythonOrJobOptions(startItem, maxitems, path, optstype)  # pylint: disable=no-member

  #############################################################################
  def getLimitedInformations(self, startItem, maxitems, path):
    """statistics."""
    return self.getManager().getLimitedInformations(startItem, maxitems, path)  # pylint: disable=no-member

  #############################################################################
  def getProcessingPassSteps(self, in_dict):
    """step."""
    return self.getManager().getProcessingPassSteps(in_dict)  # pylint: disable=no-member

  #############################################################################
  def getMoreProductionInformations(self, prodid):
    """production details."""
    return self.getManager().getMoreProductionInformations(prodid)  # pylint: disable=no-member

  #############################################################################
  def getAvailableProductions(self):
    """available productions."""
    return self.getManager().getAvailableProductions()  # pylint: disable=no-member

  #############################################################################
  def getFileHistory(self, lfn):
    """"file history."""
    return self.getManager().getFileHistory(lfn)  # pylint: disable=no-member

  #############################################################################
  def getCurrentParameter(self):
    """curent bookkeeping path."""
    return self.getManager().getCurrentParameter()  # pylint: disable=no-member

  #############################################################################
  def getQueriesTypes(self):
    """type of the current query."""
    return self.getManager().getQueriesTypes()  # pylint: disable=no-member

  #############################################################################
  def getProductionProcessingPassSteps(self, in_dict):
    """the steps which produced a given production."""
    return self.getManager().getProductionProcessingPassSteps(in_dict)  # pylint: disable=no-member

  #############################################################################
  def getAvailableDataQuality(self):
    """available data quality."""
    return self.getManager().getAvailableDataQuality()  # pylint: disable=no-member

  #############################################################################
  def setDataQualities(self, values):
    """set data qualities."""
    self.getManager().setDataQualities(values)  # pylint: disable=no-member

  #############################################################################
  def getStepsMetadata(self, bkDict):
    """returns detailed step metadata."""
    return self.getManager().getStepsMetadata(bkDict)  # pylint: disable=no-member

  #############################################################################
  def setFileTypes(self, fileTypeList):
    """it sets the file types."""
    return self.getManager().setFileTypes(fileTypeList)  # pylint: disable=no-member

  #############################################################################
  def getFilesWithMetadata(self, dataset):
    """it sets the file types.

    :param dict dataset: it is a bookkeeping dictionary, which contains the conditions used to retreive the lfns
    :return: S_OK lfns with metadata
    """
    return self.getManager().getFilesWithMetadata(dataset)  # pylint: disable=no-member



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for Workflow Modules utilities."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

import unittest
import os

from mock import patch

from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers import Resources

from LHCbDIRAC.Workflow.Modules.ModulesUtilities import lowerExtension, getEventsToProduce,\
    getCPUNormalizationFactorAvg, getProductionParameterValue


class ModulesUtilitiesTestCase(unittest.TestCase):
  """Base class for the ModulesUtilities test cases."""

  def setUp(self):

    gLogger.setLevel('DEBUG')
    self.maxDiff = None

  def tearDown(self):
    for fileProd in ['foo.txt', 'BAR.txt', 'FooBAR.ext.txt']:
      try:
        os.remove(fileProd)
      except OSError:
        continue


#############################################################################
# ModulesUtilities.py
#############################################################################

class ModulesUtilitiesSuccess(ModulesUtilitiesTestCase):

  #################################################

  def test_lowerExtension(self):

    open('foo.tXt', 'w').close()
    open('BAR.txt', 'w').close()
    open('FooBAR.eXT.TXT', 'w').close()

    lowerExtension()

    self.assertTrue('foo.txt' in os.listdir('.'))
    self.assertTrue('BAR.txt' in os.listdir('.'))
    self.assertTrue('FooBAR.ext.txt' in os.listdir('.'))

  #################################################

  def test_getEventsToProduce(self):

    CPUe = 2.0
    CPUTime = 1000000.0
    CPUNormalizationFactor = 0.5

    out = getEventsToProduce(CPUe, CPUTime, CPUNormalizationFactor)
    outExp = 200000
    self.assertEqual(out, outExp)

    out = getEventsToProduce(CPUe, CPUTime, CPUNormalizationFactor, maxNumberOfEvents=1000)
    outExp = 1000
    self.assertEqual(out, outExp)

    out = getEventsToProduce(CPUe, CPUTime, CPUNormalizationFactor, jobMaxCPUTime=100000)
    outExp = 20000
    self.assertEqual(out, outExp)

    out = getEventsToProduce(CPUe, CPUTime, CPUNormalizationFactor, maxNumberOfEvents=1000, jobMaxCPUTime=100000)
    outExp = 1000
    self.assertEqual(out, outExp)

  #################################################

  def test_getCPUNormalizationFactorAvg(self):

    with patch.object(gConfig, 'getSections') as mockGetSections:  # @UndefinedVariable
      with patch.object(Resources, 'getQueues') as mockGetQueues:  # @UndefinedVariable

        # gConfig.getSection error
        mockGetSections.return_value = S_ERROR()
        self.assertRaises(RuntimeError, getCPUNormalizationFactorAvg)

        # Resources.getQueues error
        mockGetSections.return_value = S_OK(['LCG.CERN.ch'])
        mockGetQueues.return_value = S_ERROR()
        self.assertRaises(RuntimeError, getCPUNormalizationFactorAvg)

        # no queues
        mockGetQueues.return_value = S_OK({'LCG.CERN.ch': {}})
        self.assertRaises(RuntimeError, getCPUNormalizationFactorAvg)

        # success
        mockGetQueues.return_value = S_OK({'LCG.CERN.ch': {
            'ce201.cern.ch': {'CEType': 'CREAM',
                              'OS': 'ScientificCERNSLC_Boron_5.5',
                              'Pilot': 'True',
                              'Queues': {'cream-lsf-grid_2nh_lhcb': {'MaxTotalJobs': '1000',
                                                                     'MaxWaitingJobs': '20',
                                                                     'SI00': '1000',
                                                                     'maxCPUTime': '120'},
                                         'cream-lsf-grid_lhcb': {'MaxTotalJobs': '1000',
                                                                 'MaxWaitingJobs': '100',
                                                                 'SI00': '1000',
                                                                 'WaitingToRunningRatio': '0.2',
                                                                 'maxCPUTime': '10080'}
                                         },
                                        'SI00': '5242',
                                        'SubmissionMode': 'Direct',
                                        'architecture': 'x86_64',
                                        'wnTmpDir': '.'},
            'ce202.cern.ch': {'CEType': 'CREAM',
                              'OS': 'ScientificCERNSLC_Boron_5.8',
                              'Pilot': 'True',
                              'Queues': {'cream-lsf-grid_2nh_lhcb': {'MaxTotalJobs': '1000',
                                                                     'MaxWaitingJobs': '20',
                                                                     'SI00': '1000',
                                                                     'maxCPUTime': '120'},
                                         'cream-lsf-grid_lhcb': {'MaxTotalJobs': '1000',
                                                                 'MaxWaitingJobs': '100',
                                                                 'SI00': '1000',
                                                                 'WaitingToRunningRatio': '0.2',
                                                                 'maxCPUTime': '10080'}},
                              'SI00': '5242',
                              'SubmissionMode': 'Direct',
                              'architecture': 'x86_64',
                              'wnTmpDir': '.'}}})
        out = getCPUNormalizationFactorAvg()
        self.assertEqual(out, 4.0)

  #################################################

  def test_getProductionParameterValue(self):

    emptyXML = '<Workflow></Workflow>'
    noValueProductionXML = '''
      <Workflow>
        <origin></origin>
        <description><![CDATA[prodDescription]]></description>
        <descr_short></descr_short>
        <version>0.0</version>
        <type></type>
        <name>Request_12416_MCSimulation_Sim08a/Digi13/Trig0x40760037/Reco14a/Stripping20r1NoPrescalingFlagged_EventType_13296003__1</name>
        <Parameter name="JobType" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="User specified type">
          <value></value>
        </Parameter>
      </Workflow>
    '''  # noqa
    productionXML = '''
      <Workflow>
        <origin></origin>
        <description><![CDATA[prodDescription]]></description>
        <descr_short></descr_short>
        <version>0.0</version>
        <type></type>
        <name>Request_12416_MCSimulation_Sim08a/Digi13/Trig0x40760037/Reco14a/Stripping20r1NoPrescalingFlagged_EventType_13296003__1</name>
        <Parameter name="JobType" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="User specified type">
          <value><![CDATA[MCSimulation]]></value>
        </Parameter>
      </Workflow>
    '''  # noqa

    parameterName = 'JobType'

    valueExp = 'MCSimulation'

    value = getProductionParameterValue(emptyXML, parameterName)
    self.assertEqual(value, None)

    value = getProductionParameterValue(noValueProductionXML, parameterName)
    self.assertEqual(value, None)

    value = getProductionParameterValue(productionXML, parameterName)
    self.assertEqual(value, valueExp)


#############################################################################
# Test Suite run
#############################################################################

if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(ModulesUtilitiesTestCase)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ModulesUtilitiesSuccess))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""stores the job configuration."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


class JobConfiguration:

  """JobConfiguration class."""
  #############################################################################

  def __init__(self):
    """initialize the class members."""
    self.configName_ = ""  # None
    self.configVersion_ = ""  # None
    self.date_ = ""  # None
    self.time_ = ""  # None

  #############################################################################
  def setConfigName(self, name):
    """sets the configuration name."""
    self.configName_ = name

  #############################################################################
  def getConfigName(self):
    """returns the configuration name."""
    return self.configName_

  #############################################################################
  def setConfigVersion(self, version):
    """sets the configuration version."""
    self.configVersion_ = version

  #############################################################################
  def getConfigVersion(self):
    """returns the configuration version."""
    return self.configVersion_

  #############################################################################
  def setDate(self, date):
    """sets the creation data."""
    self.date_ = date

  #############################################################################
  def getDate(self):
    """returns the creation data."""
    return self.date_

  #############################################################################
  def setTime(self, time):
    """sets the creation time."""
    self.time_ = time

  #############################################################################
  def getTime(self):
    """returns the creation time."""
    return self.time_

  #############################################################################
  def __repr__(self):
    """formats the output of the print."""
    result = 'JobConfiguration: \n'
    result += 'ConfigName:' + self.configName_ + '\n'
    result += 'ConfigVersion:' + self.configVersion_ + '\n'
    result += 'Date and Time:' + self.date_ + ' ' + self.time_
    return result

  def writeToXML(self):
    """creates an xml string."""
    result = '<Job ConfigName="' + self.getConfigName() + \
        '" ConfigVersion="' + self.getConfigVersion() + \
        '" Date="' + self.getDate() + \
        '" Time="' + self.getTime() + '">\n'
    return result

  #############################################################################



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Client module to deal with transformations, but mostly dedicated to
DataManipulation (e.g.: replications)"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.TransformationSystem.Client.Transformation import Transformation as DIRACTransformation
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup

from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

COMPONENT_NAME = 'Transformation'


class Transformation(DIRACTransformation):
  """Class for dealing with Transformation objects."""

  #############################################################################

  def __init__(self, transID=0, transClientIn=None):
    """Just params setting.

    transClient is passed here as LHCbDIRAC TransformationsClient, it
    will be self.transClient
    """

    if not transClientIn:
      self.transClient = TransformationClient()
    else:
      self.transClient = transClientIn

    super(Transformation, self).__init__(transID=transID, transClient=self.transClient)

  #############################################################################

  def testBkQuery(self, bkQuery, printOutput=False, bkClient=None):
    """just pretty print of the result of a BK Query."""

    if bkClient is None:
      bkClient = BookkeepingClient()

    res = bkClient.getFiles(bkQuery)
    if not res['OK']:
      return self._errorReport(res, 'Failed to perform BK query')
    gLogger.info('The supplied query returned %d files' % len(res['Value']))
    if printOutput:
      self._prettyPrint(res)
    return S_OK(res['Value'])

  #############################################################################

  def setBkQuery(self, queryDict, test=False):
    """set a BKK Query."""
    if test:
      res = self.testBkQuery(queryDict)
      if not res['OK']:
        return res
    transID = self.paramValues['TransformationID']
    if self.exists and transID:
      res = self.transClient.addBookkeepingQuery(transID, queryDict)
      if not res['OK']:
        return res
    self.item_called = 'BkQuery'
    self.paramValues[self.item_called] = queryDict
    return S_OK()

  #############################################################################

  def getBkQuery(self, printOutput=False):
    """get a BKK Query."""
    if self.paramValues['BkQuery']:
      return S_OK(self.paramValues['BkQuery'])
    res = self.__executeOperation('getBookkeepingQuery')
    if not res['OK']:
      return res
    self.item_called = 'BkQuery'
    self.paramValues[self.item_called] = res['Value']
    return S_OK(res['Value'])

  #############################################################################

  def deleteTransformationBkQuery(self):
    """delete a BKK Query."""
    transID = self.paramValues['TransformationID']
    if self.exists and transID:
      res = self.transClient.deleteTransformationBookkeepingQuery(transID)
      if not res['OK']:
        return res
    self.item_called = 'BkQuery'
    self.paramValues[self.item_called] = {}
    return S_OK()

  #############################################################################

  def addTransformation(self, addFiles=True, printOutput=False):
    """Add a transformation, using TransformationClient()"""
    res = super(Transformation, self).addTransformation(addFiles, printOutput)
    if res['OK']:
      transID = res['Value']
    else:
      return res

    bkQuery = self.paramValues.get('BkQuery')
    if bkQuery:
      res = self.setBkQuery(bkQuery)
      if not res['OK']:
        return self._errorReport(res, "Failed to set BK query")
    else:
      self.transClient.deleteTransformationParameter(transID, 'BkQuery')

    return S_OK(transID)

  def setSEParam(self, key, seList):
    return self.__setSE(key, seList)

  def setAdditionalParam(self, key, val):
    self.item_called = key
    return self.__setParam(val)

  # This is a trick to overwrite the __checkSEs method of the base class
  def _Transformation__checkSEs(self, seList):
    # This test allows to set some parameters empty
    if seList == []:
      return S_OK()
    if resolveSEGroup(seList):
      return S_OK()
    gLogger.error("Some SEs are unknown in %s" % ','.join(seList))
    return S_ERROR("Some StorageElements not known")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCb API Class.

The LHCb API exposes LHCb specific functionality in addition to the
standard DIRAC API.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import time

import six

from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites
from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite, getSitesForSE
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus

from LHCbDIRAC.Core.Utilities.File import makeGuid
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery
from LHCbDIRAC.DataManagementSystem.Client.DMScript import printDMResult
from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import getAccessURL

COMPONENT_NAME = 'DiracLHCb'


def getSiteForSE(se):
  """Get site name for the given SE."""
  result = getSitesForSE(se)
  if not result['OK']:
    return result
  if result['Value']:
    return S_OK(result['Value'][0])
  return S_OK('')


def translateBKPath(bkPath, procPassID=3):
  bk = [s for s in bkPath.split('/') if s]
  if procPassID < 0:
    return bk
  try:
    bkNodes = bk[0:procPassID]
    bkNodes.append('/' + '/'.join(bk[procPassID:-2]))
    bkNodes.append(bk[-2])
    bkNodes.append(bk[-1])
  except Exception:
    gLogger.error("Incorrect BKQuery")
    bkNodes = None
  return bkNodes


class DiracLHCb(Dirac):

  #############################################################################
  def __init__(self, withRepo=False, repoLocation='', operationsHelperIn=None):
    """Internal initialization of the DIRAC API."""

    super(DiracLHCb, self).__init__(withRepo=withRepo, repoLocation=repoLocation)
    self.tier1s = []

    if not operationsHelperIn:
      self.opsH = Operations()
    else:
      self.opsH = operationsHelperIn

    self._bkQueryTemplate = {'SimulationConditions': 'All',
                             'DataTakingConditions': 'All',
                             'ProcessingPass': 'All',
                             'FileType': 'All',
                             'EventType': 'All',
                             'ConfigName': 'All',
                             'ConfigVersion': 'All',
                             'Production': 0,
                             'StartRun': 0,
                             'EndRun': 0,
                             'DataQuality': 'All',
                             'Visible': 'Yes'}
    self._bkClient = BookkeepingClient()  # to expose all BK client methods indirectly

  #############################################################################
  def addRootFile(self, lfn, fullPath, diracSE, printOutput=False):
    """Add a Root file to Grid storage, an attempt is made to retrieve the POOL
    GUID of the file prior to upload.

       Example Usage:

       >>> print dirac.addFile('/lhcb/user/p/paterson/myRootFile.tar.gz','myFile.tar.gz','CERN-USER')
       {'OK': True, 'Value':{'Failed': {},
        'Successful': {'/lhcb/user/p/paterson/test/myRootFile.tar.gz': {'put': 64.246301889419556,
                                                                    'register': 1.1102778911590576}}}}

       @param lfn: Logical File Name (LFN)
       @type lfn: string
       @param diracSE: DIRAC SE name e.g. CERN-USER
       @type diracSE: strin
       @param printOutput: Optional flag to print result
       @type printOutput: boolean
       @return: S_OK,S_ERROR
    """
    return super(DiracLHCb, self).addFile(lfn, fullPath, diracSE,
                                          fileGuid=makeGuid(fullPath)[fullPath],
                                          printOutput=printOutput)

  def addFile(self, lfn, fullPath, diracSE, printOutput=False):  # pylint: disable=arguments-differ
    """Copy of addRootFile."""
    return super(DiracLHCb, self).addFile(lfn, fullPath, diracSE,
                                          fileGuid=makeGuid(fullPath)[fullPath],
                                          printOutput=printOutput)

  def getBKAncestors(self, lfns, depth=1, replica=True):
    """This function allows to retrieve ancestor files from the Bookkeeping.

        Example Usage:

        >>> dirac.getBKAncestors('/lhcb/data/2009/DST/00005727/0000/00005727_00000042_1.dst',2)
        {'OK': True, 'Value': ['/lhcb/data/2009/DST/00005727/0000/00005727_00000042_1.dst',
        '/lhcb/data/2009/RAW/FULL/LHCb/COLLISION09/63807/063807_0000000004.raw']}

       @param lfn: Logical File Name (LFN)
       @type lfn: string or list
       @param depth: Ancestor depth
       @type depth: integer
    """

    result = self._bkClient.getFileAncestors(lfns, depth, replica=replica)
    if not result['OK']:
      self.log.error('Could not get ancestors', result['Message'])
      return result
    ancestors = set(x['FileName'] for ancestors in result['Value']['Successful'].values() for x in ancestors)

    return S_OK(lfns + list(ancestors))

  #############################################################################
  def bkQueryRunsByDate(self, bkPath, startDate, endDate, dqFlag='All', selection='Runs'):
    """This function allows to create and perform a BK query given a supplied BK path

    The following BK path convention is expected:

    .. code-block:: none

      /<ConfigurationName>/<Configuration Version>/<Condition Description><Processing Pass>/<Event Type>/<File Type>

    so an example for 2016 collisions data would be:

    .. code-block:: none

      /LHCb/Collision09//LHCb/Collision16/Beam6500GeV-VeloClosed-MagDown/Real Data/Reco16/Stripping26/90000000/EW.DST

    The startDate and endDate must be specified as yyyy-mm-dd.

    Runs can be selected based on their status e.g. the selection parameter
    has the following possible attributes:

     - Runs - data for all runs in the range are queried (default)
     - ProcessedRuns - data is retrieved for runs that are processed
     - NotProcessed - data is retrieved for runs that are not yet processed.

    Example Usage:

    .. code-block:: python

      >>> dirac.bkQueryRunsByDate('/LHCb/Collision16//Real Data/90000000/RAW',
                                  '2016-08-20','2016-08-22',dqFlag='OK',selection='Runs')
      {'OK': True, 'Value': [<LFN1>,<LFN2>]}

      >>> dirac.bkQueryRunsByDate('/LHCb/Collision16/Beam6500GeV-VeloClosed-MagDown/Real'
                                  'Data/Reco16/Stripping26/90000000/EW.DST',
                                  '2016-08-20','2016-08-22',dqFlag='OK',selection='Runs')

    @param bkPath: BK path as described above
    @type bkPath: string
    @param dqFlag: Optional Data Quality flag
    @type dqFlag: string
    @param startDate: Start date  yyyy-mm-dd
    @param startDate: string
    @param endDate: End date  yyyy-mm-dd
    @param endDate: string
    @param selection: Either Runs, ProcessedRuns or NotProcessed
    @param selection: string
    @return: S_OK,S_ERROR
    """
    runSelection = ['Runs', 'ProcessedRuns', 'NotProcessed']
    if selection not in runSelection:
      return S_ERROR('Expected one of %s not "%s" for selection' % (', '.join(runSelection), selection))

    if not isinstance(bkPath, str):
      return S_ERROR('Expected string for bkPath')

    # remove any double slashes, spaces must be preserved
    # remove any empty components from leading and trailing slashes
    bkQuery = BKQuery().buildBKQuery(bkPath)
    if not bkQuery:
      return S_ERROR(
          'Please provide a BK path: '
          '/<ConfigurationName>/<Configuration Version>/<Condition Description>/<Processing Pass>'
          '/<Event Type>/<File Type>')

    if not startDate or not endDate:
      return S_ERROR('Expected both start and end dates to be defined in format: yyyy-mm-dd')

    if not isinstance(startDate, str) or not isinstance(endDate, str):
      return S_ERROR('Expected yyyy-mm-dd string for start and end dates')

    if not len(startDate.split('-')) == 3 or not len(endDate.split('-')) == 3:
      return S_ERROR('Expected yyyy-mm-dd string for start and end dates')

    start = time.time()
    result = self._bkClient.getRunsForAGivenPeriod({'StartDate': startDate, 'EndDate': endDate})
    rtime = time.time() - start
    self.log.info('BK query time: %.2f sec' % rtime)
    if not result['OK']:
      self.log.info('Could not get runs with given dates from BK with result: "%s"' % result)
      return result

    if not result['Value']:
      self.log.info('No runs selected from BK for specified dates')
      return result

    if selection not in result['Value']:
      return S_ERROR('No %s runs for specified dates' % (selection))

    runs = result['Value'][selection]
    self.log.info('Found the following %s runs:\n%s' % (len(runs), ', '.join([str(i) for i in runs])))
    # temporary until we can query for a discrete list of runs
    selectedData = []
    for run in runs:
      query = bkQuery.copy()
      query['StartRun'] = run
      query['EndRun'] = run
      query['CheckRunStatus'] = True if selection in ['ProcessedRuns', 'NotProcessed'] else False
      if dqFlag:
        check = self.__checkDQFlags(dqFlag)
        if not check['OK']:
          return check
        dqFlag = check['Value']
        query['DataQuality'] = dqFlag
      start = time.time()
      result = self._bkClient.getVisibleFilesWithMetadata(query)
      rtime = time.time() - start
      self.log.info('BK query time: %.2f sec' % rtime)
      self.log.verbose(result)
      if not result['OK']:
        return result
      self.log.info('Selected %s files for run %s' % (len(result['Value']), run))
      if result['Value']['LFNs']:
        selectedData += list(result['Value']['LFNs'])

    self.log.info('Total files selected = %s' % (len(selectedData)))
    return S_OK(selectedData)

  #############################################################################
  def bkQueryRun(self, bkPath, dqFlag='All'):
    """This function allows to create and perform a BK query given a supplied
    BK path. The following BK path convention is expected:

        /<Run Number>/<Processing Pass>/<Event Type>/<File Type>

        so an example for 2009 collisions data would be:

       /63566/Real Data + RecoToDST-07/90000000/DST

       In addition users can specify a range of runs using the following convention:

       /<Run Number 1> - <Run Number 2>/<Processing Pass>/<Event Type>/<File Type>

       so extending the above example this would look like:

       /63566-63600/Real Data + RecoToDST-07/90000000/DST

       Example Usage:

       >>> dirac.bkQueryRun('/63566/Real Data/RecoToDST-07/90000000/DST')
       {'OK':True,'Value': ['/lhcb/data/2009/DST/00005842/0000/00005842_00000008_1.dst']}

       @param bkPath: BK path as described above
       @type bkPath: string
       @param dqFlag: Optional Data Quality flag
       @type dqFlag: string
       @return: S_OK,S_ERROR
    """
    if not isinstance(bkPath, str):
      return S_ERROR('Expected string for bkPath')

    # remove any double slashes, spaces must be preserved
    # remove any empty components from leading and trailing slashes
    bkPath = translateBKPath(bkPath, procPassID=1)
    if not len(bkPath) == 4:
      return S_ERROR('Expected 4 components to the BK path: /<Run Number>/<Processing Pass>/<Event Type>/<File Type>')

    runNumberString = bkPath[0].replace('--', '-').replace(' ', '')
    startRun = 0
    endRun = 0
    if '-' in runNumberString:
      runs = runNumberString.split('-')
      if len(runs) != 2:
        return S_ERROR('Could not determine run range from "%s", try "<Run 1> - <Run2>"' % (runNumberString))
      try:
        start = int(runs[0])
        end = int(runs[1])
      except Exception:
        return S_ERROR('Invalid run range: %s' % runNumberString)
      startRun = min(start, end)
      endRun = max(start, end)
    else:
      try:
        startRun = int(runNumberString)
        endRun = startRun
      except Exception:
        return S_ERROR('Invalid run number: %s' % runNumberString)

    query = self._bkQueryTemplate.copy()
    query['StartRun'] = startRun
    query['EndRun'] = endRun
    query['ProcessingPass'] = bkPath[1]
    query['EventType'] = bkPath[2]
    query['FileType'] = bkPath[3]

    if dqFlag:
      check = self.__checkDQFlags(dqFlag)
      if not check['OK']:
        return check
      dqFlag = check['Value']
      query['DataQuality'] = dqFlag

    result = self.bkQuery(query)
    self.log.verbose(result)
    return result

  #############################################################################
  def bkQueryProduction(self, bkPath, dqFlag='All'):
    """This function allows to create and perform a BK query given a supplied
    BK path. The following BK path convention is expected:

        /<ProductionID>/[<Processing Pass>/<Event Type>/]<File Type>

        so an example for 2009 collisions data would be:

       /5842/Real Data + RecoToDST-07/90000000/DST

       Note that neither the processing pass nor the event type should be necessary. So either of them can be omitted

       a data quality flag can also optionally be provided, the full list of these is available
       via the getAllDQFlags() method.

       Example Usage:

       >>> dirac.bkQueryProduction('/5842/Real Data/RecoToDST-07/90000000/DST')
       {'OK': True, 'Value': [<LFN1>,<LFN2>]}

       @param bkPath: BK path as described above
       @type bkPath: string
       @param dqFlag: Optional Data Quality flag
       @type dqFlag: string
       @return: S_OK,S_ERROR
    """
    if not isinstance(bkPath, str):
      return S_ERROR('Expected string for bkPath')

    # remove any double slashes, spaces must be preserved
    # remove any empty components from leading and trailing slashes
    bkPath = translateBKPath(bkPath, procPassID=1)
    if len(bkPath) < 2:
      return S_ERROR('Invalid bkPath: should at least contain /ProductionID/FileType')
    query = self._bkQueryTemplate.copy()
    try:
      query['Production'] = int(bkPath[0])
    except Exception:
      return S_ERROR('Invalid production ID')
    query['FileType'] = bkPath[-1]

    if dqFlag:
      check = self.__checkDQFlags(dqFlag)
      if not check['OK']:
        return check
      dqFlag = check['Value']
      query['DataQuality'] = dqFlag

    for key, val in query.items():
      if isinstance(val, six.string_types) and val.lower() == 'all':
        query.pop(key)
    result = self.bkQuery(query)
    self.log.verbose(result)
    return result

  #############################################################################
  def bkQueryPath(self, bkPath, dqFlag='All'):
    """This function allows to create and perform a BK query given a supplied
    BK path. The following BK path convention is expected:

       /<ConfigurationName>/<Configuration Version>/<Sim or Data Taking Condition>
       /<Processing Pass>/<Event Type>/<File Type>

       so an example for 2009 collsions data would be:

       /LHCb/Collision09/Beam450GeV-VeloOpen-MagDown/Real Data + RecoToDST-07/90000000/DST

       or for MC09 simulated data:

       /MC/2010/Beam3500GeV-VeloClosed-MagDown-Nu1/2010-Sim01Reco01-withTruth/27163001/DST

       a data quality flag can also optionally be provided, the full list of these is available
       via the getAllDQFlags() method.

       Example Usage:

       >>> dirac.bkQueryPath('/MC/2010/Beam3500GeV-VeloClosed-MagDown-Nu1/Sim07/Reco06-withTruth/10012004/DST')
       {'OK': True, 'Value': [<LFN1>,<LFN2>]}

       @param bkPath: BK path as described above
       @type bkPath: string
       @param dqFlag: Optional Data Quality flag
       @type dqFlag: string
       @return: S_OK,S_ERROR
    """
    if not isinstance(bkPath, str):
      return S_ERROR('Expected string for bkPath')

    # remove any double slashes, spaces must be preserved
    # remove any empty components from leading and trailing slashes
    bkPath = translateBKPath(bkPath, procPassID=3)
    if not len(bkPath) == 6:
      return S_ERROR('Expected 6 components to the BK path: '
                     '/<ConfigurationName>/<Configuration Version>/<Sim or Data Taking Condition>'
                     '/<Processing Pass>/<Event Type>/<File Type>')

    query = self._bkQueryTemplate.copy()
    query['ConfigName'] = bkPath[0]
    query['ConfigVersion'] = bkPath[1]
    query['ProcessingPass'] = bkPath[3]
    query['EventType'] = bkPath[4]
    query['FileType'] = bkPath[5]

    if dqFlag:
      check = self.__checkDQFlags(dqFlag)
      if not check['OK']:
        return check
      dqFlag = check['Value']
      query['DataQuality'] = dqFlag

    # The problem here is that we don't know if it's a sim or data taking condition,
    # assume that if configName=MC this is simulation
    if bkPath[0].lower() == 'mc':
      query['SimulationConditions'] = bkPath[2]
    else:
      query['DataTakingConditions'] = bkPath[2]

    result = self.bkQuery(query)
    self.log.verbose(result)
    return result

  #############################################################################
  def bookkeepingQuery(self, SimulationConditions='All', DataTakingConditions='All',
                       ProcessingPass='All', FileType='All', EventType='All', ConfigName='All',
                       ConfigVersion='All', ProductionID=0, DataQuality='ALL'):
    """This function will create and perform a BK query using the supplied
    arguments and return a list of LFNs.

        Example Usage:

        >>> dirac.bookkeepingQuery(ConfigName='LHCb',ConfigVersion='Collision09',
        EventType='90000000',ProcessingPass='Real Data',DataTakingConditions='Beam450GeV-VeloOpen-MagDown')
        {'OK':True,'Value':<files>}

       @param  ConfigName: BK ConfigName
       @type ConfigName: string
       @param  EventType: BK EventType
       @type EventType: string
       @param  FileType: BK FileType
       @type FileType: string
       @param  ProcessingPass: BK ProcessingPass
       @type ProcessingPass: string
       @param  ProductionID: BK ProductionID
       @type ProductionID: integer
       @param  DataQuality: BK DataQuality
       @type DataQuality: string
       @param  ConfigVersion: BK ConfigVersion
       @type ConfigVersion: string
       @param  DataTakingConditions: BK DataTakingConditions
       @type DataTakingConditions: string
       @param  SimulationConditions: BK SimulationConditions
       @type SimulationConditions: string
       @return: S_OK,S_ERROR
    """
    query = self._bkQueryTemplate.copy()
    query['SimulationConditions'] = SimulationConditions
    query['DataTakingConditions'] = DataTakingConditions
    query['ProcessingPass'] = ProcessingPass
    query['FileType'] = FileType
    query['EventType'] = EventType
    query['ConfigName'] = ConfigName
    query['ConfigVersion'] = ConfigVersion
    query['Production'] = ProductionID
    query['DataQuality'] = DataQuality
    return self.bkQuery(query)

  #############################################################################
  def bkQuery(self, bkQueryDict):
    """Developer function. Perform a query to the LHCb Bookkeeping to return a
    list of LFN(s). This method takes a BK query dictionary.

        Example Usage:

        >>> print dirac.bkQuery(query)
        {'OK':True,'Value':<files>}

       @param bkQueryDict: BK query
       @type bkQueryDict: dictionary (see bookkeepingQuery() for keys)
       @return: S_OK,S_ERROR
    """
    problematicFields = []
    # Remove the Visible flag as anyway the method is for visible files ;-)
    # bkQueryDict.setdefault( 'Visible', 'Yes' )
    for name, value in bkQueryDict.items():
      if name not in self._bkQueryTemplate:
        problematicFields.append(name)

    if problematicFields:
      msg = 'The following fields are not valid for a BK query: %s\nValid fields include: %s' % \
            (', '.join(problematicFields), ', '.join(self._bkQueryTemplate))
      return S_ERROR(msg)

    for name, value in list(bkQueryDict.items()):
      if name == "Production" or name == "EventType" or name == "StartRun" or name == "EndRun":
        if value == 0:
          del bkQueryDict[name]
        else:
          bkQueryDict[name] = str(value)
      elif name == "FileType":
        if value.lower() == "all":
          bkQueryDict[name] = 'ALL'
      else:
        if str(value).lower() == "all":
          del bkQueryDict[name]

    if 'Production' in bkQueryDict or 'StartRun' in bkQueryDict or 'EndRun' in bkQueryDict:
      self.log.verbose('Found a specific query so loosening some restrictions to prevent BK overloading')
    else:
      if 'SimulationConditions' not in bkQueryDict and 'DataTakingConditions' not in bkQueryDict:
        return S_ERROR('A Simulation or DataTaking Condition must be specified for a BK query.')
      if 'EventType' not in bkQueryDict and 'ConfigName' not in bkQueryDict and 'ConfigVersion' not in bkQueryDict:
        return S_ERROR(
            'The minimal set of BK fields for a query is: EventType, ConfigName and ConfigVersion'
            ' in addition to a Simulation or DataTaking Condition')

    self.log.verbose('Final BK query dictionary is:')
    for item in bkQueryDict.items():
      self.log.verbose('%s : %s' % item)

    start = time.time()
    result = self._bkClient.getVisibleFilesWithMetadata(bkQueryDict)
#    result = bk.getFilesWithGivenDataSets(bkQueryDict)
    rtime = time.time() - start
    self.log.info('BK query time: %.2f sec' % rtime)

    if not result['OK']:
      return S_ERROR('BK query returned an error: "%s"' % (result['Message']))

    if not result['Value']:
      return self._errorReport('No BK files selected')

    returnedFiles = len(result['Value'])
    self.log.verbose('%s files selected from the BK' % (returnedFiles))
    return result

  #############################################################################
  def __checkDQFlags(self, flags):
    """Internal function.

    Checks the provided flags against the list of possible DQ flag
    statuses from the Bookkeeping.
    """
    dqFlags = []
    if isinstance(flags, list):
      dqFlags = flags
    else:
      dqFlags = [flags]

    bkFlags = self.getAllDQFlags()
    if not bkFlags['OK']:
      return bkFlags

    final = []
    for flag in dqFlags:
      if flag.lower() == 'all':
        final.append(flag.upper())
      else:
        flag = flag.upper()
        if flag not in bkFlags['Value']:
          msg = 'Specified DQ flag "%s" is not in allowed list: %s' % (flag, ', '.join(bkFlags['Value']))
          self.log.error(msg)
          return S_ERROR(msg)
        else:
          final.append(flag)

    # when first coding this it was not possible to use a list ;)
    if len(final) == 1:
      final = final[0]

    return S_OK(final)

  #############################################################################
  def getAllDQFlags(self, printOutput=False):
    """Helper function.  Returns the list of possible DQ flag statuses from the
    Bookkeeping.

        Example Usage:

        >>> print dirac.getAllDQFlags()
        {'OK':True,'Value':<flags>}

       @param printOutput: Optional flag to print result
       @type printOutput: boolean
       @return: S_OK,S_ERROR
    """
    result = self._bkClient.getAvailableDataQuality()
    if not result['OK']:
      self.log.error('Could not obtain possible DQ flags from BK with result:\n%s' % (result))
      return result

    if printOutput:
      flags = result['Value']
      self.log.info('Possible DQ flags from BK are: %s' % (', '.join(flags)))

    return result

  #############################################################################
  def getDataByRun(self, lfns, printOutput=False):
    """Sort the supplied lfn list by run. An S_OK object will be returned
    containing a dictionary of runs and the corresponding list of LFN(s)
    associated with them.

       Example usage:

       >>> print dirac.getDataByRun(lfns)
       {'OK': True, 'Value': {<RUN>:['<LFN>','<LFN>',...], <RUN>:['<LFN>',..]}}


       @param lfns: Logical File Name(s)
       @type lfns: list
       @param printOutput: Optional flag to print result
       @type printOutput: boolean
       @return: S_OK,S_ERROR
    """
    if isinstance(lfns, str):
      lfns = [lfns.replace('LFN:', '')]
    elif isinstance(lfns, list):
      try:
        lfns = [str(lfn.replace('LFN:', '')) for lfn in lfns]
      except ValueError as x:
        return self._errorReport(str(x), 'Expected strings for LFNs')
    else:
      return self._errorReport('Expected single string or list of strings for LFN(s)')

    runDict = {}
    start = time.time()
    result = self._bkClient.getFileMetadata(lfns)
    self.log.verbose("Obtained BK file metadata in %.2f seconds" % (time.time() - start))
    if not result['OK']:
      self.log.error('Failed to get bookkeeping metadata with result "%s"' % (result['Message']))
      return result

    for lfn, metadata in result['Value']['Successful'].items():
      if 'RunNumber' in metadata:
        runNumber = metadata['RunNumber']
        runDict.setdefault(runNumber, []).append(lfn)
      else:
        self.log.warn('Could not find run number from BK for %s' % (lfn))

    if printOutput:
      self.log.notice(self.pPrint.pformat(runDict))

    return S_OK(runDict)

  #############################################################################
  def bkMetadata(self, lfns, printOutput=False):
    """Return metadata for the supplied lfn list. An S_OK object will be
    returned containing a dictionary of LFN(s) and the corresponding metadata
    associated with them.

       Example usage:

       >>> print dirac.bkMetadata(lfns)
       {'OK': True, 'Value': {<LFN>:{'<Name>':'<Value>',...},...}}

       @param lfns: Logical File Name(s)
       @type lfns: list
       @param printOutput: Optional flag to print result
       @type printOutput: boolean
       @return: S_OK,S_ERROR
    """
    if isinstance(lfns, str):
      lfns = [lfns.replace('LFN:', '')]
    elif isinstance(lfns, list):
      try:
        lfns = [str(lfn.replace('LFN:', '')) for lfn in lfns]
      except ValueError as x:
        return self._errorReport(str(x), 'Expected strings for LFNs')
    else:
      return self._errorReport('Expected single string or list of strings for LFN(s)')

    start = time.time()
    result = self._bkClient.getFileMetadata(lfns)
    self.log.verbose("Obtained BK file metadata in %.2f seconds" % (time.time() - start))
    if not result['OK']:
      self.log.error('Failed to get bookkeeping metadata with result "%s"' % (result['Message']))
      return result

    if printOutput:
      self.log.notice(self.pPrint.pformat(result['Value']))

    return result

  #############################################################################

  def lhcbProxyInit(self, *args):  # pylint: disable=no-self-use
    """just calling the dirac-proxy-init script."""
    os.system("dirac-proxy-init -o LogLevel=NOTICE -t --rfc %s" % "' '".join(args))

  #############################################################################

  def lhcbProxyInfo(self, *args):  # pylint: disable=no-self-use
    """just calling the dirac-proxy-info script."""
    os.system("dirac-proxy-info -o LogLevel=NOTICE %s" % "' '".join(args))
  #############################################################################

  def gridWeather(self, printOutput=False):
    """This method gives a snapshot of the current Grid weather from the
    perspective of the DIRAC site and SE masks.  Tier-1 sites are returned with
    more detailed information.

       Example usage:

       >>> print dirac.gridWeather()
       {'OK': True, 'Value': {{'Sites':<siteInfo>,'SEs':<seInfo>,'Tier-1s':<tierInfo>}}

       @param printOutput: Optional flag to print result
       @type printOutput: boolean
       @return: S_OK,S_ERROR
    """

    lcgSites = gConfig.getSections('/Resources/Sites/LCG')
    if not lcgSites['OK']:
      return lcgSites

    for lcgSite in lcgSites['Value']:

      tier = gConfig.getValue('/Resources/Sites/LCG/%s/MoUTierLevel' % lcgSite, 2)
      if tier in (0, 1):
        self.tier1s.append(lcgSite)

    siteInfo = self.checkSites()
    if not siteInfo['OK']:
      return siteInfo
    siteInfo = siteInfo['Value']

    seInfo = self.checkSEs()
    if not seInfo['OK']:
      return seInfo
    seInfo = seInfo['Value']

    tierSEs = {}
    for site in self.tier1s:
      tierSEs[site] = getSEsForSite(site)['Value']

    tierInfo = {}
    for site, seList in tierSEs.items():
      tierInfo[site] = {}
      for se in seList:
        if se in seInfo:
          tierSEInfo = seInfo[se]
          tierInfo[site][se] = tierSEInfo
      if site in siteInfo['AllowedSites']:
        tierInfo[site]['MaskStatus'] = 'Allowed'
      else:
        tierInfo[site]['MaskStatus'] = 'Banned'

    if printOutput:
      self.log.notice('========> Tier-1 status in DIRAC site and SE masks')
      for site in sorted(self.tier1s):
        self.log.notice('\n====> %s is %s in site mask\n' % (site, tierInfo[site]['MaskStatus']))
        self.log.notice('%s %s %s' % ('Storage Element'.ljust(25), 'Read Status'.rjust(15), 'Write Status'.rjust(15)))
        for se in sorted(tierSEs[site]):
          if se in tierInfo[site]:
            self.log.notice('%s %s %s' % (se.ljust(25),
                                          tierInfo[site][se]['ReadStatus'].rjust(15),
                                          tierInfo[site][se]['WriteStatus'].rjust(15))
                            )

      self.log.notice('\n========> Tier-2 status in DIRAC site mask\n')
      allowedSites = siteInfo['AllowedSites']
      bannedSites = siteInfo['BannedSites']
      for site in self.tier1s:
        if site in allowedSites:
          allowedSites.remove(site)
        if site in bannedSites:
          bannedSites.remove(site)
      self.log.notice(' %s sites are in the site mask, %s are banned.\n' % (len(allowedSites), len(bannedSites)))

    summary = {'Sites': siteInfo, 'SEs': seInfo, 'Tier-1s': tierInfo}
    return S_OK(summary)

  #############################################################################
  def checkSites(self, printOutput=False):  # pylint: disable=no-self-use
    """Return the list of sites in the DIRAC site mask and those which are
    banned.

       Example usage:

       >>> print dirac.checkSites()
       {'OK': True, 'Value': {'AllowedSites':['<Site>',...],'BannedSites':[]}

       @param printOutput: Optional flag to print result
       @type printOutput: boolean
       @return: S_OK,S_ERROR
    """

    res = getSites()
    if not res['OK']:
      self.log.error('Could not get list of sites from CS', res['Message'])
      return res
    totalList = res['Value']

    res = DiracAdmin().getSiteMask()
    if not res['OK']:
      return res

    sites = res['Value']
    bannedSites = []
    for site in totalList:
      if site not in sites:
        bannedSites.append(site)

    if printOutput:
      self.log.notice('\n========> Allowed Sites\n')
      self.log.notice('\n'.join(sites))
      self.log.notice('\n========> Banned Sites\n')
      self.log.notice('\n'.join(bannedSites))
      self.log.notice('\nThere is a total of %s allowed sites and %s banned sites in the system.' % (len(sites),
                                                                                                     len(bannedSites)))

    return S_OK({'AllowedSites': sites, 'BannedSites': bannedSites})

  #############################################################################
  def checkSEs(self, printOutput=False):  # pylint: disable=no-self-use
    """Check the status of read and write operations in the DIRAC SE mask.

       Example usage:

       >>> print dirac.checkSEs()
       {'OK': True, 'Value': {<LFN>:{'<Name>':'<Value>',...},...}}

       @param printOutput: Optional flag to print result
       @type printOutput: boolean
       @return: S_OK,S_ERROR
    """
    res = gConfig.getSections('/Resources/StorageElements', True)

    if not res['OK']:
      self.log.error('Failed to get storage element information', res['Message'])
      return res

    if printOutput:
      self.log.notice('%s %s %s' % ('Storage Element'.ljust(25), 'Read Status'.rjust(15), 'Write Status'.rjust(15)))

    seList = sorted(res['Value'])
    result = {}
    rss = ResourceStatus()
    for se in seList:
      res = rss.getElementStatus(se, 'StorageElement')
      if not res['OK']:
        self.log.error("Failed to get StorageElement status for %s" % se)
      else:
        readState = res['Value'].get('ReadAccess', 'Active')
        writeState = res['Value'].get('WriteAccess', 'Active')
        result[se] = {'ReadStatus': readState, 'WriteStatus': writeState}
        if printOutput:
          self.log.notice('%s %s %s' % (se.ljust(25), readState.rjust(15), writeState.rjust(15)))

    return S_OK(result)

  def splitInputDataBySize(self, lfns, maxSizePerJob=20, printOutput=False):
    """Split the supplied lfn list by the replicas present at the possible
    destination sites, based on a maximum size. An S_OK object will be returned
    containing a list of lists in order to create the jobs.

       Example usage:

       >>> d.splitInputDataBySize(lfns,10)
       {'OK': True, 'Value': [['<LFN>'], ['<LFN>']]}


       @param lfns: Logical File Name(s) to split
       @type lfns: list
       @param maxSizePerJob: Maximum size (in GB) per bunch
       @type maxSizePerJob: integer
       @param printOutput: Optional flag to print result
       @type printOutput: boolean
       @return: S_OK,S_ERROR
    """
    sitesForSE = {}
    if isinstance(lfns, str):
      lfns = [lfns.replace('LFN:', '')]
    elif isinstance(lfns, list):
      try:
        lfns = [str(lfn.replace('LFN:', '')) for lfn in lfns]
      except TypeError as x:
        return self._errorReport(str(x), 'Expected strings for LFNs')
    else:
      return self._errorReport('Expected single string or list of strings for LFN(s)')

    if not isinstance(maxSizePerJob, int):
      try:
        maxSizePerJob = int(maxSizePerJob)
      except ValueError as x:
        return self._errorReport(str(x), 'Expected integer for maxSizePerJob')
    maxSizePerJob *= 1000 * 1000 * 1000

    replicaDict = self.getReplicas(lfns)
    if not replicaDict['OK']:
      return replicaDict
    replicas = replicaDict['Value']['Successful']
    if not replicas:
      return self._errorReport(list(replicaDict['Value']['Failed'].items())[0],
                               'Failed to get replica information')
    siteLfns = {}
    for lfn, reps in replicas.items():
      possibleSites = set(site
                          for se in reps
                          for site in sitesForSE.setdefault(se, getSitesForSE(se).get('Value', [])))
      siteLfns.setdefault(','.join(sorted(possibleSites)), []).append(lfn)

    if '' in siteLfns:
      # Some files don't have active replicas
      return self._errorReport('No active replica found for', str(siteLfns['']))
    # Get size of files
    metadataDict = self.getLfnMetadata(lfns, printOutput)
    if not metadataDict['OK']:
      return metadataDict
    fileSizes = dict((lfn, metadataDict['Value']['Successful'].get(lfn, {}).get('Size', maxSizePerJob))
                     for lfn in lfns)

    lfnGroups = []
    # maxSize is in GB
    for files in siteLfns.values():
      # Now get bunches of files,
      # Sort in decreasing size
      files.sort(cmp=(lambda f1, f2: fileSizes[f2] - fileSizes[f1]))
      while files:
        # print [( lfn, fileSizes[lfn] ) for lfn in files]
        group = []
        sizeTot = 0
        for lfn in list(files):
          size = fileSizes[lfn]
          if size >= maxSizePerJob:
            lfnGroups.append([lfn])
          elif sizeTot + size < maxSizePerJob:
            sizeTot += size
            group.append(lfn)
            files.remove(lfn)
        if group:
          lfnGroups.append(group)

    if printOutput:
      self.log.notice(self.pPrint.pformat(lfnGroups))
    return S_OK(lfnGroups)

    #############################################################################

  def getAccessURL(self, lfn, storageElement, protocol=None, printOutput=False):
    """Allows to retrieve an access URL for an LFN replica given a valid DIRAC
    SE name.  Contacts the file catalog and contacts the site SRM endpoint
    behind the scenes.

       Example Usage:

       >>> print dirac.getAccessURL('/lhcb/data/CCRC08/DST/00000151/0000/00000151_00004848_2.dst','CERN-RAW')
       {'OK': True, 'Value': {'Successful': {'srm://...': {'SRM2': 'rfio://...'}}, 'Failed': {}}}

       :param lfn: Logical File Name (LFN)
       :type lfn: str or python:list
       :param storageElement: DIRAC SE name e.g. CERN-RAW
       :type storageElement: string
       :param printOutput: Optional flag to print result
       :type printOutput: boolean
       :returns: S_OK,S_ERROR
    """
    ret = self._checkFileArgument(lfn, 'LFN')
    if not ret['OK']:
      return ret
    lfn = ret['Value']
    if isinstance(lfn, six.string_types):
      lfn = [lfn]
    results = getAccessURL(lfn, storageElement, protocol=protocol)
    if printOutput:
      printDMResult(results, empty="File not at SE", script="dirac-dms-lfn-accessURL")
    return results

  #############################################################################

  def _getLocalInputData(self, parameters):
    """LHCb extension of DIRAC API's _getLocalInputData.

    Only used for handling ancestors.
    """
    inputData = parameters.get('InputData')
    if inputData:
      self.log.debug("DiracLHCb._getLocalInputData. InputData: %s" % inputData)
      if isinstance(inputData, six.string_types):
        inputData = inputData.split(';')
      inputData = [lfn.strip('LFN:') for lfn in inputData]
      ancestorsDepth = int(parameters.get('AncestorDepth', 0))
      if ancestorsDepth:
        self.log.debug("DiracLHCb._getLocalInputData. ancestorsDepth: %d" % ancestorsDepth)
        res = self._bkClient.getFileAncestors(inputData, ancestorsDepth)
        if not res['OK']:
          self.log.error("Can't get ancestors", res['Message'])
          return res
        ancestorsLFNs = []
        for ancestorsLFN in res['Value']['Successful'].values():
          ancestorsLFNs += [i['FileName'] for i in ancestorsLFN]
        self.log.info("DiracLHCb._getLocalInputData: adding %d ancestors" % len(ancestorsLFNs))
        self.log.verbose("%s", ', '.join(ancestorsLFNs))
        inputData += ancestorsLFNs

    return S_OK(inputData)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""  The MC Simulation Template creates workflows for the following simulation
     use-cases:
      WORKFLOW1: Simulation+Selection+Merge
      WORKFLOW2: Simulation+Selection+MCMerge
      WORKFLOW3: Simulation+Selection
      WORKFLOW4: Simulation+MCMerge
      WORKFLOW5: Simulation

    Exotic things you might want to do:
    * run a local test:
      - of the MC: just set the localTestFlag to True
      - of the merging/stripping: set pr.prodsToLaunch to, e.g., [2], and adjust the pr.inputs at the end of the script
    * run only part of the request on the Grid:
      - for the MC: just set pr.prodsToLaunch = [1]
      - for the merge and/or stripping: set pr.prodsToLaunch, then set pr.previousProdID
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from six.moves import range

import ast

from DIRAC.Core.Base import Script
Script.parseCommandLine()

from DIRAC import gConfig, gLogger, exit as DIRACexit
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import ProductionRequest

__RCSID__ = "$Id$"

#
# Link compression level to the output file visibility. If any file types is visible, the corresponding
# step takes the maximum compression level
# Input : dicts including the visibility of output files per step
# Output: modified list of compression levels
#


def modifyCompression(dict1, dict2, default):

  compression = []

  combinedVis = dict1.copy()
  combinedVis.update(dict2)
  for stepN in sorted(combinedVis):
    if combinedVis[stepN] == 'Y':
      compression.append('HIGH')
    else:
      compression.append(default)

  return compression

#
# Simple "check" in case the visibility includes only one element
# Input : dict of visibility
# Output: modified dict of visibility
#


def fillVisList(vdict, num):

  # Assuming that, if there's only one element in the list of output visibility flags, every step will catch that flag
  if len(vdict) == 1:
    val = vdict[list(vdict)[0]]
    vdict = dict([(str(i), val) for i in range(int(list(vdict)[0]), int(list(vdict)[0]) + num)])
  # Another assumption: if the number of steps is bigger than that of vis flags,
  # then extend the list with the last flag available
  # to fill the "holes"
  # if len(vlist) < len(slist):
  #  vlist.extend( vlist[-1] * (len(slist) - len(vlist)) )

  return vdict


gLogger = gLogger.getSubLogger('MCSimulation_run.py')
currentSetup = gConfig.getValue('DIRAC/Setup')

pr = ProductionRequest()

stepsList = ['{{p1Step}}']
stepsList.append('{{p2Step}}')
stepsList.append('{{p3Step}}')
stepsList.append('{{p4Step}}')
stepsList.append('{{p5Step}}')
stepsList.append('{{p6Step}}')
stepsList.append('{{p7Step}}')
stepsList.append('{{p8Step}}')
stepsList.append('{{p9Step}}')
stepsList.append('{{p10Step}}')
stepsList.append('{{p11Step}}')
stepsList.append('{{p12Step}}')
stepsList.append('{{p13Step}}')
stepsList.append('{{p14Step}}')
stepsList.append('{{p15Step}}')
stepsList.append('{{p16Step}}')
stepsList.append('{{p17Step}}')
stepsList.append('{{p18Step}}')
stepsList.append('{{p19Step}}')
stepsList.append('{{p20Step}}')
pr.stepsList = stepsList

###########################################
# Configurable and fixed parameters
###########################################

pr.appendName = '{{WorkflowAppendName#GENERAL: Workflow string to append to production name#1}}'

w = '{{w#----->WORKFLOW: choose one below#}}'
w1 = '{{w1#-WORKFLOW1: Simulation#False}}'
w2 = '{{w2#-WORKFLOW2: Simulation(Gauss) + AllOthers#False}}'
w3 = '{{w3#-WORKFLOW3: Simulation(Gauss) + AllOthers + Merge#False}}'

localTestFlag = '{{localTestFlag#GENERAL: Set True for local test#False}}'
validationFlag = '{{validationFlag#GENERAL: Set True for validation prod - will create histograms#False}}'

pr.configName = '{{BKConfigName#GENERAL: BK configuration name e.g. MC #MC}}'
extraOptions = '{{extraOptions#GENERAL: extra options as python dict stepID:options#}}'

targets = '{{Target#PROD-1:MC: Target for MC (e.g. Tier2, ALL, LCG.CERN.cern or BAN:site1:site2#ALL}}'
eventsPerJob = '{{eventsPerJob#PROD-1:MC: Number of events per job#-1}}'
MCPriority = '{{MCPriority#PROD-1:MC: Production priority#0}}'
MCmulticoreFlag = '{{MCMulticoreFLag#PROD-1:MC multicore flag#True}}'
MCNumberOfProcessors = '{{MCNumberOfProcessors#PROD-1:MC jobs min/max n of processors#0,0}}'
MCSimulationType = '{{MCSimulationType#PROD-1:MC: type of MCSimulation#MCSimulation}}'
pr.mcTesting = ast.literal_eval('{{MCTesting#PROD-1:MC: testing enabled#True}}')
simulationCompressionLvl = '{{simulationCompressionLvl#PROD-1: Compression level#LOW}}'

simulationOutputVisFlag = ast.literal_eval(
    '{{simulationOutputVisFlag#PROD-1: Simulation visib flag dictionary (one flag per step {"step":"Y|N"})#{} }}')
try:
  simVisFlagSpecial = '{{simulationOutputVisFlagSpecial'
  simVisFlagSpecial += '#PROD-1: Special Visibility flag of output files'
  simVisFlagSpecial += ' (a dictionary {"step n":{("FType":flag)}} )#}}'
  simulationOutputVisFlagSpecial = ast.literal_eval(simVisFlagSpecial)
except SyntaxError:
  simulationOutputVisFlagSpecial = {}

selectionPlugin = '{{selectionPlugin#PROD-2:Selection: plugin e.g. Standard, BySize#BySize}}'
selectionGroupSize = '{{selectionGroupSize#PROD-2:Selection: input files total size (we\'ll download)#5}}'
selectionPriority = '{{selectionPriority#PROD-2:Selection: Job Priority e.g. 8 by default#8}}'
selectionCPU = '{{selectionCPU#PROD-2:Selection: Max CPU time in secs#100000}}'
removeInputSelection = '{{removeInputSelection#PROD-2:Selection: remove inputs#True}}'
selmulticoreFlag = '{{selMulticoreFLag#PROD-2:Selection: multicore flag#True}}'
selNumberOfProcessors = '{{selNumberOfProcessors#PROD-2:Selection jobs min/max n of processors#0,0}}'
selectionCompressionLvl = '{{selectionCompressionLvl#PROD-2:Selection: Compression level#LOW}}'
selectionOutputVisFlag = ast.literal_eval(
    '{{selectionOutputVisFlag#PROD-2: Selection visibility flag dictionary ({"step n": "Y|N"})#{} }}')
try:
  selVisFlagSpecial = '{{selectionOutputVisFlagSpecial'
  selVisFlagSpecial += '#PROD-2: Special Visibility flag of output files (a dictionary {"step n":{"FType":flag}} )#}}'
  selectionOutputVisFlagSpecial = ast.literal_eval(selVisFlagSpecial)
except SyntaxError:
  selectionOutputVisFlagSpecial = {}

mergingPlugin = '{{MergingPlugin#PROD-3:Merging: plugin e.g. Standard, BySize#BySize}}'
mergingGroupSize = '{{MergingGroupSize#PROD-3:Merging: Group Size e.g. BySize = GB file size#5}}'
mergingPriority = '{{MergingPriority#PROD-3:Merging: Job Priority e.g. 8 by default#8}}'
mergingCPU = '{{mergingCPU#PROD-3:Merging: Max CPU time in secs#100000}}'
removeInputMerge = '{{removeInputMerge#PROD-3:Merging: remove inputs#True}}'
mergemulticoreFlag = '{{mergeMulticoreFLag#PROD-3:Merging: multicore flag#True}}'
mergeNumberOfProcessors = '{{mergeNumberOfProcessors#PROD-3:Merging jobs min/max n of processors#0,0}}'
mergeCompressionLvl = '{{mergeCompressionLvl#PROD-3:Merging: Compression level#HIGH}}'
mergeOutputVisFlag = ast.literal_eval(
    '{{mergeOutputVisFlag#PROD-3: Merge visibility flag dictionary ({"step":"Y|N"})#{} }}')
try:
  mergeVisFlagSpecial = '{{mergeOutputVisFlagSpecial'
  mergeVisFlagSpecial += '#PROD-3: Special Visibility flag of output files'
  mergeVisFlagSpecial += '(a dictionary {"step n":{"FType":flag}} )#}}'
  mergeOutputVisFlagSpecial = ast.literal_eval(mergeVisFlagSpecial)
except SyntaxError:
  mergeOutputVisFlagSpecial = {}

pr.configVersion = '{{mcConfigVersion}}'
pr.eventType = '{{eventType}}'
# Often MC requests are defined with many subrequests but we want to retain
# the parent ID for viewing on the production monitoring page. If a parent
# request is defined then this is used.
pr.parentRequestID = '{{_parent}}'
pr.requestID = '{{ID}}'

if extraOptions:
  pr.extraOptions = ast.literal_eval(extraOptions)
pr.prodGroup = '{{pDsc}}'
pr.dataTakingConditions = '{{simDesc}}'

MCPriority = int(MCPriority)
selectionPriority = int(selectionPriority)
mergingPriority = int(mergingPriority)

removeInputMerge = ast.literal_eval(removeInputMerge)
removeInputSelection = ast.literal_eval(removeInputSelection)

MCNumberOfProcessors = tuple(int(x) for x in MCNumberOfProcessors.replace(' ', '').split(','))
selNumberOfProcessors = tuple(int(x) for x in selNumberOfProcessors.replace(' ', '').split(','))
mergeNumberOfProcessors = tuple(int(x) for x in mergeNumberOfProcessors.replace(' ', '').split(','))

###########################################
# LHCb conventions implied by the above
###########################################

localTestFlag = ast.literal_eval(localTestFlag)
validationFlag = ast.literal_eval(validationFlag)

if localTestFlag:
  pr.testFlag = True
  pr.publishFlag = False
  pr.prodsToLaunch = [1]

pr.outConfigName = pr.configName

w1 = ast.literal_eval(w1)
w2 = ast.literal_eval(w2)
w3 = ast.literal_eval(w3)

if not w1 and not w2 and not w3:
  gLogger.error('Vladimir, I told you to select at least one workflow!')
  DIRACexit(2)

elif w1:
  pr.prodsTypeList = [MCSimulationType]
  pr.outputSEs = ['Tier1_MC-DST']

  pr.stepsInProds = [list(range(1, len(pr.stepsList) + 1))]
  pr.removeInputsFlags = [False]
  pr.priorities = [MCPriority]
  pr.cpus = [100000]
  pr.outputFileSteps = [str(len(pr.stepsList))]
  pr.targets = [targets]
  pr.events = [eventsPerJob]
  pr.groupSizes = [1]
  pr.plugins = ['']
  pr.inputDataPolicies = ['']
  pr.bkQueries = ['']
  pr.multicore = [MCmulticoreFlag]
  pr.processors = [MCNumberOfProcessors]

  simulationOutputVisFlag = fillVisList(simulationOutputVisFlag, len(pr.stepsInProds[0]))
  pr.compressionLvl = modifyCompression(simulationOutputVisFlag,
                                        simulationOutputVisFlagSpecial,
                                        simulationCompressionLvl)

  pr.outputVisFlag = [simulationOutputVisFlag]
  pr.specialOutputVisFlag = [simulationOutputVisFlagSpecial]

  # pr.resolveSteps()

elif w2:
  pr.prodsTypeList = [MCSimulationType, 'MCReconstruction']
  pr.outputSEs = ['Tier1-Buffer', 'Tier1_MC-DST']

  pr.stepsInProds = [[1, ], list(range(2, len(pr.stepsList) + 1))]
  pr.outputFileSteps = [str(len(pr.stepsInProds[0])),
                        str(len(pr.stepsInProds[1]))]

  pr.removeInputsFlags = [False, removeInputSelection]
  pr.priorities = [MCPriority, selectionPriority]
  pr.cpus = [100000, selectionCPU]
  pr.targets = [targets, '']
  pr.events = [eventsPerJob, -1]
  pr.groupSizes = [1, selectionGroupSize]
  pr.plugins = ['', selectionPlugin]
  pr.inputDataPolicies = ['', 'download']
  pr.bkQueries = ['', 'fromPreviousProd']
  pr.multicore = [MCmulticoreFlag, selmulticoreFlag]
  pr.processors = [MCNumberOfProcessors, selNumberOfProcessors]

  pr.compressionLvl = [simulationCompressionLvl] * len(pr.stepsInProds[0]) + \
                      [selectionCompressionLvl] * len(pr.stepsInProds[1])

  simulationOutputVisFlag = fillVisList(simulationOutputVisFlag, len(pr.stepsInProds[0]))
  selectionOutputVisFlag = fillVisList(selectionOutputVisFlag, len(pr.stepsInProds[1]))

  pr.compressionLvl = modifyCompression(simulationOutputVisFlag,
                                        simulationOutputVisFlagSpecial,
                                        simulationCompressionLvl)
  pr.compressionLvl += modifyCompression(selectionOutputVisFlag,
                                         selectionOutputVisFlagSpecial,
                                         selectionCompressionLvl)

  pr.outputVisFlag = [simulationOutputVisFlag, selectionOutputVisFlag]
  pr.specialOutputVisFlag = [simulationOutputVisFlagSpecial, selectionOutputVisFlagSpecial]

  # pr.resolveSteps()

elif w3:
  pr.prodsTypeList = [MCSimulationType, 'MCReconstruction', 'MCMerge']
  pr.outputSEs = ['Tier1-Buffer', 'Tier1-Buffer', 'Tier1_MC-DST']

  pr.stepsInProds = [[1, ], list(range(2, len(pr.stepsList))), [len(pr.stepsList)]]
  pr.outputFileSteps = ['1', str(len(pr.stepsInProds[1])), '1']

  pr.removeInputsFlags = [False, removeInputSelection, removeInputMerge]
  pr.priorities = [MCPriority, selectionPriority, mergingPriority]
  pr.cpus = [100000, selectionCPU, mergingCPU]
  pr.targets = [targets, '', '']
  pr.events = [eventsPerJob, -1, -1]
  pr.groupSizes = [1, selectionGroupSize, mergingGroupSize]
  pr.plugins = ['', selectionPlugin, mergingPlugin]
  pr.inputDataPolicies = ['', 'download', 'download']
  pr.bkQueries = ['', 'fromPreviousProd', 'fromPreviousProd']
  pr.multicore = [MCmulticoreFlag, selmulticoreFlag, mergemulticoreFlag]
  pr.processors = [MCNumberOfProcessors, selNumberOfProcessors, mergeNumberOfProcessors]

  simulationOutputVisFlag = fillVisList(simulationOutputVisFlag, len(pr.stepsInProds[0]))
  selectionOutputVisFlag = fillVisList(selectionOutputVisFlag, len(pr.stepsInProds[1]))
  mergeOutputVisFlag = fillVisList(mergeOutputVisFlag, len(pr.stepsInProds[2]))

  pr.compressionLvl = modifyCompression(simulationOutputVisFlag,
                                        simulationOutputVisFlagSpecial,
                                        simulationCompressionLvl)
  pr.compressionLvl += modifyCompression(selectionOutputVisFlag,
                                         selectionOutputVisFlagSpecial,
                                         selectionCompressionLvl)
  pr.compressionLvl += modifyCompression(mergeOutputVisFlag,
                                         mergeOutputVisFlagSpecial,
                                         mergeCompressionLvl)

  pr.outputVisFlag = [simulationOutputVisFlag, selectionOutputVisFlag, mergeOutputVisFlag]
  pr.specialOutputVisFlag = [simulationOutputVisFlagSpecial, selectionOutputVisFlagSpecial, mergeOutputVisFlagSpecial]

# In case we want just to test, we publish in the certification/test part of the BKK
if currentSetup == 'LHCb-Certification' or pr.testFlag:
  pr.outConfigName = 'certification'
  pr.configVersion = 'test'

if pr.testFlag:
  pr.extend = '10'
  mergingGroupSize = '1'
  MCCpu = '50000'
  pr.previousProdID = 0  # set this for, e.g., launching only merging

# Validation implies few things, like saving all the outputs, and adding a GAUSSHIST
if validationFlag:
  pr.resolveSteps()
  pr.outConfigName = 'validation'
  # Adding GAUSSHIST to the list of outputs to be produced (by the first step, which is Gauss)
  if 'GAUSSHIST' not in pr.stepsListDict[0]['fileTypesOut']:
    pr.stepsListDict[0]['fileTypesOut'].append('GAUSSHIST')
  pr.outputFileSteps = [''] * len(pr.prodsTypeList)


res = pr.buildAndLaunchRequest()
if not res['OK']:
  gLogger.error("Errors with submission: %s" % res['Message'])
  DIRACexit(2)
else:
  gLogger.always("Submitted %s" % str(res['Value']))



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Simple extension of base class."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.TransformationSystem.Agent.ValidateOutputDataAgent import ValidateOutputDataAgent \
    as DIRACValidateOutputDataAgent

from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient
from LHCbDIRAC.DataManagementSystem.Client.DataIntegrityClient import DataIntegrityClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

AGENT_NAME = 'Transformation/ValidateOutputDataAgent'


class ValidateOutputDataAgent(DIRACValidateOutputDataAgent):
  """Simple extension of base class."""

  def __init__(self, *args, **kwargs):
    """c'tor."""
    DIRACValidateOutputDataAgent.__init__(self, *args, **kwargs)

    self.integrityClient = None
    self.fileCatalog = None
    self.transClient = None
    self.storageUsageClient = None

  def initialize(self):
    """standard initialize method for DIRAC agents."""
    res = DIRACValidateOutputDataAgent.initialize(self)
    if not res['OK']:
      return res

    self.integrityClient = DataIntegrityClient()
    self.fileCatalog = FileCatalog()
    self.transClient = TransformationClient()
    self.storageUsageClient = StorageUsageClient()

    return S_OK()

  def checkTransformationIntegrity(self, prodID):
    """This method contains the real work."""
    gLogger.info("-" * 40)
    gLogger.info("Checking the integrity of production %s" % prodID)
    gLogger.info("-" * 40)

    res = self.getTransformationDirectories(prodID)
    if not res['OK']:
      return res
    directories = res['Value']

    ######################################################
    #
    # This check performs BK->Catalog->SE
    #
    res = self.integrityClient.productionToCatalog(prodID)
    if not res['OK']:
      gLogger.error(res['Message'])
      return res
    bk2catalogMetadata = res['Value']['CatalogMetadata']
    bk2catalogReplicas = res['Value']['CatalogReplicas']
    res = self.integrityClient.checkPhysicalFiles(bk2catalogReplicas, bk2catalogMetadata)
    if not res['OK']:
      gLogger.error(res['Message'])
      return res

    if not directories:
      return S_OK()

    ######################################################
    #
    # This check performs Catalog->BK and Catalog->SE for possible output directories
    #
    res = self.fileCatalog.exists(directories)
    if not res['OK']:
      gLogger.error(res['Message'])
      return res
    for directory, error in res['Value']['Failed']:
      gLogger.error('Failed to determine existance of directory', '%s %s' % (directory, error))
    if res['Value']['Failed']:
      return S_ERROR("Failed to determine the existance of directories")
    directoryExists = res['Value']['Successful']
    for directory in sorted(directoryExists):
      if not directoryExists[directory]:
        continue
      iRes = self.integrityClient.catalogDirectoryToBK(directory)
      if not iRes['OK']:
        gLogger.error(iRes['Message'])
        return iRes
      catalogDirMetadata = iRes['Value']['CatalogMetadata']
      catalogDirReplicas = iRes['Value']['CatalogReplicas']
      catalogMetadata = {}
      catalogReplicas = {}
      for lfn in catalogDirMetadata:
        if lfn not in bk2catalogMetadata:
          catalogMetadata[lfn] = catalogDirMetadata[lfn]
          if lfn in catalogDirReplicas:
            catalogReplicas[lfn] = catalogDirReplicas[lfn]
      if not catalogMetadata:
        continue
      res = self.integrityClient.checkPhysicalFiles(catalogReplicas, catalogMetadata)
      if not res['OK']:
        gLogger.error(res['Message'])
        return res

    return S_OK()

  def getTransformationDirectories(self, transID):
    """get the directories for the supplied transformation from the
    transformation system.

    :param self: self reference
    :param int transID: transformation ID
    """

    res = DIRACValidateOutputDataAgent.getTransformationDirectories(self, transID)

    if res['OK']:
      directories = res['Value']
    else:
      return res

    if 'StorageUsage' in self.directoryLocations:
      res = self.storageUsageClient.getStorageDirectories('', '', transID, [])
      if not res['OK']:
        self.log.error("Failed to obtain storage usage directories", res['Message'])
        return res
      transDirectories = res['Value']
      directories = self._addDirs(transID, transDirectories, directories)

    if not directories:
      self.log.info("No output directories found")
    directories = sorted(directories)
    return S_OK(directories)
